# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import sys
import os

import math
import logging
import threading
import queue
from collections import defaultdict

from pyrockoeost.guts import Object, Int, List, Tuple, String, Timestamp, Dict
from pyrockoeost import util, trace
from pyrockoeost.progress import progress
from pyrockoeost.plot import nice_time_tick_inc_approx_secs

from . import model, io, cache, dataset

from .model import to_kind_id, WaveformOrder, to_kind, to_codes, \
    STATION, CHANNEL, RESPONSE, EVENT, WAVEFORM, codes_patterns_list, \
    codes_patterns_for_kind
from .client import fdsn, catalog
from .selection import Selection, filldocs
from .database import abspath
from .operators.base import Operator, CodesPatternFiltering
from . import client, environment, error

logger = logging.getLogger('psq.base')

guts_prefix = 'squirrel'


def nonef(f, xs):
    xs_ = [x for x in xs if x is not None]
    if xs_:
        return f(xs_)
    else:
        return None


def make_task(*args):
    return progress.task(*args, logger=logger)


def lpick(condition, seq):
    ft = [], []
    for ele in seq:
        ft[int(bool(condition(ele)))].append(ele)

    return ft


def len_plural(obj):
    return len(obj), '' if len(obj) == 1 else 's'


def blocks(tmin, tmax, deltat, nsamples_block=100000):
    tblock = nice_time_tick_inc_approx_secs(
        util.to_time_float(deltat * nsamples_block))
    iblock_min = int(math.floor(tmin / tblock))
    iblock_max = int(math.ceil(tmax / tblock))
    for iblock in range(iblock_min, iblock_max):
        yield iblock * tblock, (iblock+1) * tblock


def gaps(avail, tmin, tmax):
    assert tmin < tmax

    data = [(tmax, 1), (tmin, -1)]
    for (tmin_a, tmax_a) in avail:
        assert tmin_a < tmax_a
        data.append((tmin_a, 1))
        data.append((tmax_a, -1))

    data.sort()
    s = 1
    gaps = []
    tmin_g = None
    for t, x in data:
        if s == 1 and x == -1:
            tmin_g = t
        elif s == 0 and x == 1 and tmin_g is not None:
            tmax_g = t
            if tmin_g != tmax_g:
                gaps.append((tmin_g, tmax_g))

        s += x

    return gaps


def order_key(order):
    return (order.codes, order.tmin, order.tmax)


def _is_exact(pat):
    return not ('*' in pat or '?' in pat or ']' in pat or '[' in pat)


def prefix_tree(tups):
    if not tups:
        return []

    if len(tups[0]) == 1:
        return sorted((tup[0], []) for tup in tups)

    d = defaultdict(list)
    for tup in tups:
        d[tup[0]].append(tup[1:])

    sub = []
    for k in sorted(d.keys()):
        sub.append((k, prefix_tree(d[k])))

    return sub


def match_time_span(tmin, tmax, obj):
    return (obj.tmin is None or tmax is None or obj.tmin <= tmax) \
        and (tmin is None or obj.tmax is None or tmin < obj.tmax)


class Batch(object):
    '''
    Batch of waveforms from window-wise data extraction.

    Encapsulates state and results yielded for each window in window-wise
    waveform extraction with the :py:meth:`Squirrel.chopper_waveforms` method.

    *Attributes:*

    .. py:attribute:: tmin

        Start of this time window.

    .. py:attribute:: tmax

        End of this time window.

    .. py:attribute:: i

        Index of this time window in sequence.

    .. py:attribute:: n

        Total number of time windows in sequence.

    .. py:attribute:: igroup

        Index of this time window's sequence group.

    .. py:attribute:: ngroups

        Total number of sequence groups.

    .. py:attribute:: traces

        Extracted waveforms for this time window.
    '''

    def __init__(self, tmin, tmax, i, n, igroup, ngroups, traces):
        self.tmin = tmin
        self.tmax = tmax
        self.i = i
        self.n = n
        self.igroup = igroup
        self.ngroups = ngroups
        self.traces = traces


class Squirrel(Selection):
    '''
    Prompt, lazy, indexing, caching, dynamic seismological dataset access.

    :param env:
        Squirrel environment instance or directory path to use as starting
        point for its detection. By default, the current directory is used as
        starting point. When searching for a usable environment the directory
        ``'.squirrel'`` or ``'squirrel'`` in the current (or starting point)
        directory is used if it exists, otherwise the parent directories are
        search upwards for the existence of such a directory. If no such
        directory is found, the user's global Squirrel environment
        ``'$HOME/.pyrockoeost/squirrel'`` is used.
    :type env:
        :py:class:`~pyrockoeost.squirrel.environment.Environment` or
        :py:class:`str`

    :param database:
        Database instance or path to database. By default the
        database found in the detected Squirrel environment is used.
    :type database:
        :py:class:`~pyrockoeost.squirrel.database.Database` or :py:class:`str`

    :param cache_path:
        Directory path to use for data caching. By default, the ``'cache'``
        directory in the detected Squirrel environment is used.
    :type cache_path:
        :py:class:`str`

    :param persistent:
        If given a name, create a persistent selection.
    :type persistent:
        :py:class:`str`

    This is the central class of the Squirrel framework. It provides a unified
    interface to query and access seismic waveforms, station meta-data and
    event information from local file collections and remote data sources. For
    prompt responses, a profound database setup is used under the hood. To
    speed up assemblage of ad-hoc data selections, files are indexed on first
    use and the extracted meta-data is remembered in the database for
    subsequent accesses. Bulk data is lazily loaded from disk and remote
    sources, just when requested. Once loaded, data is cached in memory to
    expedite typical access patterns. Files and data sources can be dynamically
    added to and removed from the Squirrel selection at runtime.

    Queries are restricted to the contents of the files currently added to the
    Squirrel selection (usually a subset of the file meta-information
    collection in the database). This list of files is referred to here as the
    "selection". By default, temporary tables are created in the attached
    database to hold the names of the files in the selection as well as various
    indices and counters. These tables are only visible inside the application
    which created them and are deleted when the database connection is closed
    or the application exits. To create a selection which is not deleted at
    exit, supply a name to the ``persistent`` argument of the Squirrel
    constructor. Persistent selections are shared among applications using the
    same database.

    **Method summary**

    Some of the methods are implemented in :py:class:`Squirrel`'s base class
    :py:class:`~pyrockoeost.squirrel.selection.Selection`.

    .. autosummary::

        ~Squirrel.add
        ~Squirrel.add_source
        ~Squirrel.add_fdsn
        ~Squirrel.add_catalog
        ~Squirrel.add_dataset
        ~Squirrel.add_virtual
        ~Squirrel.update
        ~Squirrel.update_waveform_promises
        ~Squirrel.advance_accessor
        ~Squirrel.clear_accessor
        ~Squirrel.reload
        ~pyrockoeost.squirrel.selection.Selection.iter_paths
        ~Squirrel.iter_nuts
        ~Squirrel.iter_kinds
        ~Squirrel.iter_deltats
        ~Squirrel.iter_codes
        ~pyrockoeost.squirrel.selection.Selection.get_paths
        ~Squirrel.get_nuts
        ~Squirrel.get_kinds
        ~Squirrel.get_deltats
        ~Squirrel.get_codes
        ~Squirrel.get_counts
        ~Squirrel.get_time_span
        ~Squirrel.get_deltat_span
        ~Squirrel.get_nfiles
        ~Squirrel.get_nnuts
        ~Squirrel.get_total_size
        ~Squirrel.get_stats
        ~Squirrel.get_content
        ~Squirrel.get_stations
        ~Squirrel.get_channels
        ~Squirrel.get_responses
        ~Squirrel.get_events
        ~Squirrel.get_waveform_nuts
        ~Squirrel.get_waveforms
        ~Squirrel.chopper_waveforms
        ~Squirrel.get_coverage
        ~Squirrel.pile
        ~Squirrel.snuffle
        ~Squirrel.glob_codes
        ~pyrockoeost.squirrel.selection.Selection.get_database
        ~Squirrel.print_tables
    '''

    def __init__(
            self, env=None, database=None, cache_path=None, persistent=None):

        if not isinstance(env, environment.Environment):
            env = environment.get_environment(env)

        if database is None:
            database = env.expand_path(env.database_path)

        if cache_path is None:
            cache_path = env.expand_path(env.cache_path)

        if persistent is None:
            persistent = env.persistent

        Selection.__init__(
            self, database=database, persistent=persistent)

        self.get_database().set_basepath(os.path.dirname(env.get_basepath()))

        self._content_caches = {
            'waveform': cache.ContentCache(),
            'default': cache.ContentCache()}

        self._cache_path = cache_path

        self._sources = []
        self._operators = []
        self._operator_registry = {}

        self._pending_orders = []

        self._pile = None
        self._n_choppers_active = 0

        self._names.update({
            'nuts': self.name + '_nuts',
            'kind_codes_count': self.name + '_kind_codes_count',
            'coverage': self.name + '_coverage'})

        with self.transaction('create tables') as cursor:
            self._create_tables_squirrel(cursor)

    def _create_tables_squirrel(self, cursor):

        cursor.execute(self._register_table(self._sql(
            '''
                CREATE TABLE IF NOT EXISTS %(db)s.%(nuts)s (
                    nut_id integer PRIMARY KEY,
                    file_id integer,
                    file_segment integer,
                    file_element integer,
                    kind_id integer,
                    kind_codes_id integer,
                    tmin_seconds integer,
                    tmin_offset integer,
                    tmax_seconds integer,
                    tmax_offset integer,
                    kscale integer)
            ''')))

        cursor.execute(self._register_table(self._sql(
            '''
                CREATE TABLE IF NOT EXISTS %(db)s.%(kind_codes_count)s (
                    kind_codes_id integer PRIMARY KEY,
                    count integer)
            ''')))

        cursor.execute(self._sql(
            '''
                CREATE UNIQUE INDEX IF NOT EXISTS %(db)s.%(nuts)s_file_element
                    ON %(nuts)s (file_id, file_segment, file_element)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE INDEX IF NOT EXISTS %(db)s.%(nuts)s_index_file_id
                ON %(nuts)s (file_id)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE INDEX IF NOT EXISTS %(db)s.%(nuts)s_index_tmin_seconds
                ON %(nuts)s (kind_id, tmin_seconds)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE INDEX IF NOT EXISTS %(db)s.%(nuts)s_index_tmax_seconds
                ON %(nuts)s (kind_id, tmax_seconds)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE INDEX IF NOT EXISTS %(db)s.%(nuts)s_index_kscale
                ON %(nuts)s (kind_id, kscale, tmin_seconds)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_delete_nuts
                BEFORE DELETE ON main.files FOR EACH ROW
                BEGIN
                  DELETE FROM %(nuts)s WHERE file_id == old.file_id;
                END
            '''))

        # trigger only on size to make silent update of mtime possible
        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_delete_nuts2
                BEFORE UPDATE OF size ON main.files FOR EACH ROW
                BEGIN
                  DELETE FROM %(nuts)s WHERE file_id == old.file_id;
                END
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS
                    %(db)s.%(file_states)s_delete_files
                BEFORE DELETE ON %(db)s.%(file_states)s FOR EACH ROW
                BEGIN
                    DELETE FROM %(nuts)s WHERE file_id == old.file_id;
                END
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_inc_kind_codes
                BEFORE INSERT ON %(nuts)s FOR EACH ROW
                BEGIN
                    INSERT OR IGNORE INTO %(kind_codes_count)s VALUES
                    (new.kind_codes_id, 0);
                    UPDATE %(kind_codes_count)s
                    SET count = count + 1
                    WHERE new.kind_codes_id
                        == %(kind_codes_count)s.kind_codes_id;
                END
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_dec_kind_codes
                BEFORE DELETE ON %(nuts)s FOR EACH ROW
                BEGIN
                    UPDATE %(kind_codes_count)s
                    SET count = count - 1
                    WHERE old.kind_codes_id
                        == %(kind_codes_count)s.kind_codes_id;
                END
            '''))

        cursor.execute(self._register_table(self._sql(
            '''
                CREATE TABLE IF NOT EXISTS %(db)s.%(coverage)s (
                    kind_codes_id integer,
                    time_seconds integer,
                    time_offset integer,
                    step integer)
            ''')))

        cursor.execute(self._sql(
            '''
                CREATE UNIQUE INDEX IF NOT EXISTS %(db)s.%(coverage)s_time
                    ON %(coverage)s (kind_codes_id, time_seconds, time_offset)
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_add_coverage
                AFTER INSERT ON %(nuts)s FOR EACH ROW
                BEGIN
                    INSERT OR IGNORE INTO %(coverage)s VALUES
                    (new.kind_codes_id, new.tmin_seconds, new.tmin_offset, 0)
                    ;
                    UPDATE %(coverage)s
                    SET step = step + 1
                    WHERE new.kind_codes_id == %(coverage)s.kind_codes_id
                        AND new.tmin_seconds == %(coverage)s.time_seconds
                        AND new.tmin_offset == %(coverage)s.time_offset
                    ;
                    INSERT OR IGNORE INTO %(coverage)s VALUES
                    (new.kind_codes_id, new.tmax_seconds, new.tmax_offset, 0)
                    ;
                    UPDATE %(coverage)s
                    SET step = step - 1
                    WHERE new.kind_codes_id == %(coverage)s.kind_codes_id
                        AND new.tmax_seconds == %(coverage)s.time_seconds
                        AND new.tmax_offset == %(coverage)s.time_offset
                    ;
                    DELETE FROM %(coverage)s
                        WHERE new.kind_codes_id == %(coverage)s.kind_codes_id
                            AND new.tmin_seconds == %(coverage)s.time_seconds
                            AND new.tmin_offset == %(coverage)s.time_offset
                            AND step == 0
                    ;
                    DELETE FROM %(coverage)s
                        WHERE new.kind_codes_id == %(coverage)s.kind_codes_id
                            AND new.tmax_seconds == %(coverage)s.time_seconds
                            AND new.tmax_offset == %(coverage)s.time_offset
                            AND step == 0
                    ;
                END
            '''))

        cursor.execute(self._sql(
            '''
                CREATE TRIGGER IF NOT EXISTS %(db)s.%(nuts)s_remove_coverage
                BEFORE DELETE ON %(nuts)s FOR EACH ROW
                BEGIN
                    INSERT OR IGNORE INTO %(coverage)s VALUES
                    (old.kind_codes_id, old.tmin_seconds, old.tmin_offset, 0)
                    ;
                    UPDATE %(coverage)s
                    SET step = step - 1
                    WHERE old.kind_codes_id == %(coverage)s.kind_codes_id
                        AND old.tmin_seconds == %(coverage)s.time_seconds
                        AND old.tmin_offset == %(coverage)s.time_offset
                    ;
                    INSERT OR IGNORE INTO %(coverage)s VALUES
                    (old.kind_codes_id, old.tmax_seconds, old.tmax_offset, 0)
                    ;
                    UPDATE %(coverage)s
                    SET step = step + 1
                    WHERE old.kind_codes_id == %(coverage)s.kind_codes_id
                        AND old.tmax_seconds == %(coverage)s.time_seconds
                        AND old.tmax_offset == %(coverage)s.time_offset
                    ;
                    DELETE FROM %(coverage)s
                        WHERE old.kind_codes_id == %(coverage)s.kind_codes_id
                            AND old.tmin_seconds == %(coverage)s.time_seconds
                            AND old.tmin_offset == %(coverage)s.time_offset
                            AND step == 0
                    ;
                    DELETE FROM %(coverage)s
                        WHERE old.kind_codes_id == %(coverage)s.kind_codes_id
                            AND old.tmax_seconds == %(coverage)s.time_seconds
                            AND old.tmax_offset == %(coverage)s.time_offset
                            AND step == 0
                    ;
                END
            '''))

    def _delete(self):
        '''Delete database tables associated with this Squirrel.'''

        with self.transaction('delete tables') as cursor:
            for s in '''
                    DROP TRIGGER %(db)s.%(nuts)s_delete_nuts;
                    DROP TRIGGER %(db)s.%(nuts)s_delete_nuts2;
                    DROP TRIGGER %(db)s.%(file_states)s_delete_files;
                    DROP TRIGGER %(db)s.%(nuts)s_inc_kind_codes;
                    DROP TRIGGER %(db)s.%(nuts)s_dec_kind_codes;
                    DROP TABLE %(db)s.%(nuts)s;
                    DROP TABLE %(db)s.%(kind_codes_count)s;
                    DROP TRIGGER IF EXISTS %(db)s.%(nuts)s_add_coverage;
                    DROP TRIGGER IF EXISTS %(db)s.%(nuts)s_remove_coverage;
                    DROP TABLE IF EXISTS %(db)s.%(coverage)s;
                    '''.strip().splitlines():

                cursor.execute(self._sql(s))

        Selection._delete(self)

    @filldocs
    def add(self,
            paths,
            kinds=None,
            format='detect',
            include=None,
            exclude=None,
            check=True):

        '''
        Add files to the selection.

        :param paths:
            Iterator yielding paths to files or directories to be added to the
            selection. Recurses into directories. If given a ``str``, it
            is treated as a single path to be added.
        :type paths:
            :py:class:`list` of :py:class:`str`

        :param kinds:
            Content types to be made available through the Squirrel selection.
            By default, all known content types are accepted.
        :type kinds:
            :py:class:`list` of :py:class:`str`

        :param format:
            File format identifier or ``'detect'`` to enable auto-detection
            (available: %(file_formats)s).
        :type format:
            str

        :param include:
            If not ``None``, files are only included if their paths match the
            given regular expression pattern.
        :type format:
            str

        :param exclude:
            If not ``None``, files are only included if their paths do not
            match the given regular expression pattern.
        :type format:
            str

        :param check:
            If ``True``, all file modification times are checked to see if
            cached information has to be updated (slow). If ``False``, only
            previously unknown files are indexed and cached information is used
            for known files, regardless of file state (fast, corrresponds to
            Squirrel's ``--optimistic`` mode). File deletions will go
            undetected in the latter case.
        :type check:
            bool

        :Complexity:
            O(log N)
        '''

        if isinstance(kinds, str):
            kinds = (kinds,)

        if isinstance(paths, str):
            paths = [paths]

        kind_mask = model.to_kind_mask(kinds)

        with progress.view():
            Selection.add(
                self, util.iter_select_files(
                    paths,
                    show_progress=False,
                    include=include,
                    exclude=exclude,
                    pass_through=lambda path: path.startswith('virtual:')
                ), kind_mask, format)

            self._load(check)
            self._update_nuts()

    def reload(self):
        '''
        Check for modifications and reindex modified files.

        Based on file modification times.
        '''

        self._set_file_states_force_check()
        self._load(check=True)
        self._update_nuts()

    def add_virtual(self, nuts, virtual_paths=None):
        '''
        Add content which is not backed by files.

        :param nuts:
            Content pieces to be added.
        :type nuts:
            iterator yielding :py:class:`~pyrockoeost.squirrel.model.Nut` objects

        :param virtual_paths:
            List of virtual paths to prevent creating a temporary list of the
            nuts while aggregating the file paths for the selection.
        :type virtual_paths:
            :py:class:`list` of :py:class:`str`

        Stores to the main database and the selection.
        '''

        if isinstance(virtual_paths, str):
            virtual_paths = [virtual_paths]

        if virtual_paths is None:
            if not isinstance(nuts, list):
                nuts = list(nuts)
            virtual_paths = set(nut.file_path for nut in nuts)

        Selection.add(self, virtual_paths)
        self.get_database().dig(nuts)
        self._update_nuts()

    def add_volatile(self, nuts):
        if not isinstance(nuts, list):
            nuts = list(nuts)

        paths = list(set(nut.file_path for nut in nuts))
        io.backends.virtual.add_nuts(nuts)
        self.add_virtual(nuts, paths)
        self._volatile_paths.extend(paths)

    def add_volatile_waveforms(self, traces):
        '''
        Add in-memory waveforms which will be removed when the app closes.
        '''

        name = model.random_name()

        path = 'virtual:volatile:%s' % name

        nuts = []
        for itr, tr in enumerate(traces):
            assert tr.tmin <= tr.tmax
            tmin_seconds, tmin_offset = model.tsplit(tr.tmin)
            tmax_seconds, tmax_offset = model.tsplit(
                tr.tmin + tr.data_len()*tr.deltat)

            nuts.append(model.Nut(
                file_path=path,
                file_format='virtual',
                file_segment=itr,
                file_element=0,
                file_mtime=0,
                codes=tr.codes,
                tmin_seconds=tmin_seconds,
                tmin_offset=tmin_offset,
                tmax_seconds=tmax_seconds,
                tmax_offset=tmax_offset,
                deltat=tr.deltat,
                kind_id=to_kind_id('waveform'),
                content=tr))

        self.add_volatile(nuts)
        return path

    def _load(self, check):
        for _ in io.iload(
                self,
                content=[],
                skip_unchanged=True,
                check=check):
            pass

    def _update_nuts(self, transaction=None):
        transaction = transaction or self.transaction('update nuts')
        with make_task('Aggregating selection') as task, \
                transaction as cursor:

            self._conn.set_progress_handler(task.update, 100000)
            nrows = cursor.execute(self._sql(
                '''
                    INSERT INTO %(db)s.%(nuts)s
                    SELECT NULL,
                        nuts.file_id, nuts.file_segment, nuts.file_element,
                        nuts.kind_id, nuts.kind_codes_id,
                        nuts.tmin_seconds, nuts.tmin_offset,
                        nuts.tmax_seconds, nuts.tmax_offset,
                        nuts.kscale
                    FROM %(db)s.%(file_states)s
                    INNER JOIN nuts
                        ON %(db)s.%(file_states)s.file_id == nuts.file_id
                    INNER JOIN kind_codes
                        ON nuts.kind_codes_id ==
                           kind_codes.kind_codes_id
                    WHERE %(db)s.%(file_states)s.file_state != 2
                        AND (((1 << kind_codes.kind_id)
                            & %(db)s.%(file_states)s.kind_mask) != 0)
                ''')).rowcount

            task.update(nrows)
            self._set_file_states_known(transaction)
            self._conn.set_progress_handler(None, 0)

    def add_source(self, source, check=True):
        '''
        Add remote resource.

        :param source:
            Remote data access client instance.
        :type source:
           subclass of :py:class:`~pyrockoeost.squirrel.client.base.Source`
        '''

        self._sources.append(source)
        source.setup(self, check=check)

    def add_fdsn(self, *args, **kwargs):
        '''
        Add FDSN site for transparent remote data access.

        Arguments are passed to
        :py:class:`~pyrockoeost.squirrel.client.fdsn.FDSNSource`.
        '''

        self.add_source(fdsn.FDSNSource(*args, **kwargs))

    def add_catalog(self, *args, **kwargs):
        '''
        Add online catalog for transparent event data access.

        Arguments are passed to
        :py:class:`~pyrockoeost.squirrel.client.catalog.CatalogSource`.
        '''

        self.add_source(catalog.CatalogSource(*args, **kwargs))

    def add_dataset(self, ds, check=True):
        '''
        Read dataset description from file and add its contents.

        :param ds:
            Path to dataset description file or dataset description object
            . See :py:mod:`~pyrockoeost.squirrel.dataset`.
        :type ds:
            :py:class:`str` or :py:class:`~pyrockoeost.squirrel.dataset.Dataset`

        :param check:
            If ``True``, all file modification times are checked to see if
            cached information has to be updated (slow). If ``False``, only
            previously unknown files are indexed and cached information is used
            for known files, regardless of file state (fast, corrresponds to
            Squirrel's ``--optimistic`` mode). File deletions will go
            undetected in the latter case.
        :type check:
            bool
        '''
        if isinstance(ds, str):
            ds = dataset.read_dataset(ds)

        ds.setup(self, check=check)

    def _get_selection_args(
            self, kind_id,
            obj=None, tmin=None, tmax=None, time=None, codes=None):

        if codes is not None:
            codes = codes_patterns_for_kind(kind_id, codes)

        if time is not None:
            tmin = time
            tmax = time

        if obj is not None:
            tmin = tmin if tmin is not None else obj.tmin
            tmax = tmax if tmax is not None else obj.tmax
            codes = codes if codes is not None else codes_patterns_for_kind(
                kind_id, obj.codes)

        return tmin, tmax, codes

    def _get_selection_args_str(self, *args, **kwargs):

        tmin, tmax, codes = self._get_selection_args(*args, **kwargs)
        return 'tmin: %s, tmax: %s, codes: %s' % (
            util.time_to_str(tmin) if tmin is not None else 'none',
            util.time_to_str(tmax) if tmax is not None else 'none',
            ','.join(str(entry) for entry in codes))

    def _selection_args_to_kwargs(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None):

        return dict(obj=obj, tmin=tmin, tmax=tmax, time=time, codes=codes)

    def _timerange_sql(self, tmin, tmax, kind, cond, args, naiv):

        tmin_seconds, tmin_offset = model.tsplit(tmin)
        tmax_seconds, tmax_offset = model.tsplit(tmax)
        if naiv:
            cond.append('%(db)s.%(nuts)s.tmin_seconds <= ?')
            args.append(tmax_seconds)
        else:
            tscale_edges = model.tscale_edges
            tmin_cond = []
            for kscale in range(tscale_edges.size + 1):
                if kscale != tscale_edges.size:
                    tscale = int(tscale_edges[kscale])
                    tmin_cond.append('''
                        (%(db)s.%(nuts)s.kind_id = ?
                         AND %(db)s.%(nuts)s.kscale == ?
                         AND %(db)s.%(nuts)s.tmin_seconds BETWEEN ? AND ?)
                    ''')
                    args.extend(
                        (to_kind_id(kind), kscale,
                         tmin_seconds - tscale - 1, tmax_seconds + 1))

                else:
                    tmin_cond.append('''
                        (%(db)s.%(nuts)s.kind_id == ?
                         AND %(db)s.%(nuts)s.kscale == ?
                         AND %(db)s.%(nuts)s.tmin_seconds <= ?)
                    ''')

                    args.extend(
                        (to_kind_id(kind), kscale, tmax_seconds + 1))
            if tmin_cond:
                cond.append(' ( ' + ' OR '.join(tmin_cond) + ' ) ')

        cond.append('%(db)s.%(nuts)s.tmax_seconds >= ?')
        args.append(tmin_seconds)

    def _codes_match_sql(self, kind_id, codes, cond, args):
        pats = codes_patterns_for_kind(kind_id, codes)
        if pats is None:
            return

        pats_exact = []
        pats_nonexact = []
        for pat in pats:
            spat = pat.safe_str
            (pats_exact if _is_exact(spat) else pats_nonexact).append(spat)

        cond_exact = None
        if pats_exact:
            cond_exact = ' ( kind_codes.codes IN ( %s ) ) ' % ', '.join(
                '?'*len(pats_exact))

            args.extend(pats_exact)

        cond_nonexact = None
        if pats_nonexact:
            cond_nonexact = ' ( %s ) ' % ' OR '.join(
                    ('kind_codes.codes GLOB ?',) * len(pats_nonexact))

            args.extend(pats_nonexact)

        if cond_exact and cond_nonexact:
            cond.append(' ( %s OR %s ) ' % (cond_exact, cond_nonexact))

        elif cond_exact:
            cond.append(cond_exact)

        elif cond_nonexact:
            cond.append(cond_nonexact)

    def iter_nuts(
            self, kind=None, tmin=None, tmax=None, codes=None, naiv=False,
            kind_codes_ids=None, path=None, limit=None):

        '''
        Iterate over content entities matching given constraints.

        :param kind:
            Content kind (or kinds) to extract.
        :type kind:
            :py:class:`str`, :py:class:`list` of :py:class:`str`

        :param tmin:
            Start time of query interval.
        :type tmin:
            timestamp

        :param tmax:
            End time of query interval.
        :type tmax:
            timestamp

        :param codes:
            List of code patterns to query.
        :type codes:
            :py:class:`list` of :py:class:`~pyrockoeost.squirrel.model.Codes`
            objects appropriate for the queried content type, or anything which
            can be converted to such objects.

        :param naiv:
            Bypass time span lookup through indices (slow, for testing).
        :type naiv:
            :py:class:`bool`

        :param kind_codes_ids:
            Kind-codes IDs of contents to be retrieved (internal use).
        :type kind_codes_ids:
            :py:class:`list` of :py:class:`int`

        :yields:
            :py:class:`~pyrockoeost.squirrel.model.Nut` objects representing the
            intersecting content.

        :complexity:
            O(log N) for the time selection part due to heavy use of database
            indices.

        Query time span is treated as a half-open interval ``[tmin, tmax)``.
        However, if ``tmin`` equals ``tmax``, the edge logics are modified to
        closed-interval so that content intersecting with the time instant ``t
        = tmin = tmax`` is returned (otherwise nothing would be returned as
        ``[t, t)`` never matches anything).

        Time spans of content entities to be matched are also treated as half
        open intervals, e.g. content span ``[0, 1)`` is matched by query span
        ``[0, 1)`` but not by ``[-1, 0)`` or ``[1, 2)``. Also here, logics are
        modified to closed-interval when the content time span is an empty
        interval, i.e. to indicate a time instant. E.g. time instant 0 is
        matched by ``[0, 1)`` but not by ``[-1, 0)`` or ``[1, 2)``.
        '''

        if not isinstance(kind, str):
            if kind is None:
                kind = model.g_content_kinds
            for kind_ in kind:
                for nut in self.iter_nuts(kind_, tmin, tmax, codes):
                    yield nut

            return

        kind_id = to_kind_id(kind)

        cond = []
        args = []
        if tmin is not None or tmax is not None:
            assert kind is not None
            if tmin is None:
                tmin = self.get_time_span()[0]
            if tmax is None:
                tmax = self.get_time_span()[1] + 1.0

            self._timerange_sql(tmin, tmax, kind, cond, args, naiv)

        cond.append('kind_codes.kind_id == ?')
        args.append(kind_id)

        if codes is not None:
            self._codes_match_sql(kind_id, codes, cond, args)

        if kind_codes_ids is not None:
            cond.append(
                ' ( kind_codes.kind_codes_id IN ( %s ) ) ' % ', '.join(
                    '?'*len(kind_codes_ids)))

            args.extend(kind_codes_ids)

        db = self.get_database()
        if path is not None:
            cond.append('files.path == ?')
            args.append(db.relpath(abspath(path)))

        sql = ('''
            SELECT
                files.path,
                files.format,
                files.mtime,
                files.size,
                %(db)s.%(nuts)s.file_segment,
                %(db)s.%(nuts)s.file_element,
                kind_codes.kind_id,
                kind_codes.codes,
                %(db)s.%(nuts)s.tmin_seconds,
                %(db)s.%(nuts)s.tmin_offset,
                %(db)s.%(nuts)s.tmax_seconds,
                %(db)s.%(nuts)s.tmax_offset,
                kind_codes.deltat
            FROM files
            INNER JOIN %(db)s.%(nuts)s
                ON files.file_id == %(db)s.%(nuts)s.file_id
            INNER JOIN kind_codes
                ON %(db)s.%(nuts)s.kind_codes_id == kind_codes.kind_codes_id
            ''')

        if cond:
            sql += ''' WHERE ''' + ' AND '.join(cond)

        if limit is not None:
            sql += ''' LIMIT %i''' % limit

        sql = self._sql(sql)
        if tmin is None and tmax is None:
            for row in self._conn.execute(sql, args):
                row = (db.abspath(row[0]),) + row[1:]
                nut = model.Nut(values_nocheck=row)
                yield nut
        else:
            assert tmin is not None and tmax is not None
            if tmin == tmax:
                for row in self._conn.execute(sql, args):
                    row = (db.abspath(row[0]),) + row[1:]
                    nut = model.Nut(values_nocheck=row)
                    if (nut.tmin <= tmin < nut.tmax) \
                            or (nut.tmin == nut.tmax and tmin == nut.tmin):

                        yield nut
            else:
                for row in self._conn.execute(sql, args):
                    row = (db.abspath(row[0]),) + row[1:]
                    nut = model.Nut(values_nocheck=row)
                    if (tmin < nut.tmax and nut.tmin < tmax) \
                            or (nut.tmin == nut.tmax
                                and tmin <= nut.tmin < tmax):

                        yield nut

    def get_nuts(self, *args, **kwargs):
        '''
        Get content entities matching given constraints.

        Like :py:meth:`iter_nuts` but returns results as a list.
        '''

        return list(self.iter_nuts(*args, **kwargs))

    def _split_nuts(
            self, kind, tmin=None, tmax=None, codes=None, path=None):

        kind_id = to_kind_id(kind)
        tmin_seconds, tmin_offset = model.tsplit(tmin)
        tmax_seconds, tmax_offset = model.tsplit(tmax)

        names_main_nuts = dict(self._names)
        names_main_nuts.update(db='main', nuts='nuts')

        db = self.get_database()

        def main_nuts(s):
            return s % names_main_nuts

        with self.transaction('split nuts') as cursor:
            # modify selection and main
            for sql_subst in [
                    self._sql, main_nuts]:

                cond = []
                args = []

                self._timerange_sql(tmin, tmax, kind, cond, args, False)

                if codes is not None:
                    self._codes_match_sql(kind_id, codes, cond, args)

                if path is not None:
                    cond.append('files.path == ?')
                    args.append(db.relpath(abspath(path)))

                sql = sql_subst('''
                    SELECT
                        %(db)s.%(nuts)s.nut_id,
                        %(db)s.%(nuts)s.tmin_seconds,
                        %(db)s.%(nuts)s.tmin_offset,
                        %(db)s.%(nuts)s.tmax_seconds,
                        %(db)s.%(nuts)s.tmax_offset,
                        kind_codes.deltat
                    FROM files
                    INNER JOIN %(db)s.%(nuts)s
                        ON files.file_id == %(db)s.%(nuts)s.file_id
                    INNER JOIN kind_codes
                        ON %(db)s.%(nuts)s.kind_codes_id == kind_codes.kind_codes_id
                    WHERE ''' + ' AND '.join(cond))  # noqa

                insert = []
                delete = []
                for row in cursor.execute(sql, args):
                    nut_id, nut_tmin_seconds, nut_tmin_offset, \
                        nut_tmax_seconds, nut_tmax_offset, nut_deltat = row

                    nut_tmin = model.tjoin(
                        nut_tmin_seconds, nut_tmin_offset)
                    nut_tmax = model.tjoin(
                        nut_tmax_seconds, nut_tmax_offset)

                    if nut_tmin < tmax and tmin < nut_tmax:
                        if nut_tmin < tmin:
                            insert.append((
                                nut_tmin_seconds, nut_tmin_offset,
                                tmin_seconds, tmin_offset,
                                model.tscale_to_kscale(
                                    tmin_seconds - nut_tmin_seconds),
                                nut_id))

                        if tmax < nut_tmax:
                            insert.append((
                                tmax_seconds, tmax_offset,
                                nut_tmax_seconds, nut_tmax_offset,
                                model.tscale_to_kscale(
                                    nut_tmax_seconds - tmax_seconds),
                                nut_id))

                        delete.append((nut_id,))

                sql_add = '''
                    INSERT INTO %(db)s.%(nuts)s (
                            file_id, file_segment, file_element, kind_id,
                            kind_codes_id, tmin_seconds, tmin_offset,
                            tmax_seconds, tmax_offset, kscale )
                        SELECT
                            file_id, file_segment, file_element,
                            kind_id, kind_codes_id, ?, ?, ?, ?, ?
                        FROM %(db)s.%(nuts)s
                        WHERE nut_id == ?
                '''
                cursor.executemany(sql_subst(sql_add), insert)

                sql_delete = '''
                    DELETE FROM %(db)s.%(nuts)s WHERE nut_id == ?
                '''
                cursor.executemany(sql_subst(sql_delete), delete)

    def get_time_span(self, kinds=None, tight=True, dummy_limits=True):
        '''
        Get time interval over all content in selection.

        :param kinds:
            If not ``None``, restrict query to given content kinds.
        :type kind:
            list of str

        :complexity:
            O(1), independent of the number of nuts.

        :returns:
            ``(tmin, tmax)``, combined time interval of queried content kinds.
        '''

        sql_min = self._sql('''
            SELECT MIN(tmin_seconds), MIN(tmin_offset)
            FROM %(db)s.%(nuts)s
            WHERE kind_id == ?
                AND tmin_seconds == (
                    SELECT MIN(tmin_seconds)
                    FROM %(db)s.%(nuts)s
                    WHERE kind_id == ?)
        ''')

        sql_max = self._sql('''
            SELECT MAX(tmax_seconds), MAX(tmax_offset)
            FROM %(db)s.%(nuts)s
            WHERE kind_id == ?
                AND tmax_seconds == (
                    SELECT MAX(tmax_seconds)
                    FROM %(db)s.%(nuts)s
                    WHERE kind_id == ?)
        ''')

        gtmin = None
        gtmax = None

        if isinstance(kinds, str):
            kinds = [kinds]

        if kinds is None:
            kind_ids = model.g_content_kind_ids
        else:
            kind_ids = model.to_kind_ids(kinds)

        tmins = []
        tmaxs = []
        for kind_id in kind_ids:
            for tmin_seconds, tmin_offset in self._conn.execute(
                    sql_min, (kind_id, kind_id)):
                tmins.append(model.tjoin(tmin_seconds, tmin_offset))

            for (tmax_seconds, tmax_offset) in self._conn.execute(
                    sql_max, (kind_id, kind_id)):
                tmaxs.append(model.tjoin(tmax_seconds, tmax_offset))

        tmins = [tmin if tmin != model.g_tmin else None for tmin in tmins]
        tmaxs = [tmax if tmax != model.g_tmax else None for tmax in tmaxs]

        if tight:
            gtmin = nonef(min, tmins)
            gtmax = nonef(max, tmaxs)
        else:
            gtmin = None if None in tmins else nonef(min, tmins)
            gtmax = None if None in tmaxs else nonef(max, tmaxs)

        if dummy_limits:
            if gtmin is None:
                gtmin = model.g_tmin
            if gtmax is None:
                gtmax = model.g_tmax

        return gtmin, gtmax

    def has(self, kinds):
        '''
        Check availability of given content kinds.

        :param kinds:
            Content kinds to query.
        :type kind:
            list of str

        :returns:
            ``True`` if any of the queried content kinds is available
            in the selection.
        '''
        self_tmin, self_tmax = self.get_time_span(
            kinds, dummy_limits=False)

        return None not in (self_tmin, self_tmax)

    def get_deltat_span(self, kind):
        '''
        Get min and max sampling interval of all content of given kind.

        :param kind:
            Content kind
        :type kind:
            str

        :returns: ``(deltat_min, deltat_max)``
        '''

        deltats = [
            deltat for deltat in self.get_deltats(kind)
            if deltat is not None]

        if deltats:
            return min(deltats), max(deltats)
        else:
            return None, None

    def iter_kinds(self, codes=None):
        '''
        Iterate over content types available in selection.

        :param codes:
            If given, get kinds only for selected codes identifier.
            Only a single identifier may be given here and no pattern matching
            is done, currently.
        :type codes:
            :py:class:`~pyrockoeost.squirrel.model.Codes`

        :yields:
            Available content kinds as :py:class:`str`.

        :complexity:
            O(1), independent of number of nuts.
        '''

        return self._database._iter_kinds(
            codes=codes,
            kind_codes_count='%(db)s.%(kind_codes_count)s' % self._names)

    def iter_deltats(self, kind=None):
        '''
        Iterate over sampling intervals available in selection.

        :param kind:
            If given, get sampling intervals only for a given content type.
        :type kind:
            str

        :yields:
            :py:class:`float` values.

        :complexity:
            O(1), independent of number of nuts.
        '''
        return self._database._iter_deltats(
            kind=kind,
            kind_codes_count='%(db)s.%(kind_codes_count)s' % self._names)

    def iter_codes(self, kind=None):
        '''
        Iterate over content identifier code sequences available in selection.

        :param kind:
            If given, get codes only for a given content type.
        :type kind:
            str

        :yields:
            :py:class:`tuple` of :py:class:`str`

        :complexity:
            O(1), independent of number of nuts.
        '''
        return self._database._iter_codes(
            kind=kind,
            kind_codes_count='%(db)s.%(kind_codes_count)s' % self._names)

    def _iter_codes_info(self, kind=None, codes=None):
        '''
        Iterate over number of occurrences of any (kind, codes) combination.

        :param kind:
            If given, get counts only for selected content type.
        :type kind:
            str

        :yields:
            Tuples of the form ``(kind, codes, deltat, kind_codes_id, count)``.

        :complexity:
            O(1), independent of number of nuts.
        '''
        return self._database._iter_codes_info(
            kind=kind,
            codes=codes,
            kind_codes_count='%(db)s.%(kind_codes_count)s' % self._names)

    def get_kinds(self, codes=None):
        '''
        Get content types available in selection.

        :param codes:
            If given, get kinds only for selected codes identifier.
            Only a single identifier may be given here and no pattern matching
            is done, currently.
        :type codes:
            :py:class:`~pyrockoeost.squirrel.model.Codes`

        :returns:
            Sorted list of available content types.
        :rtype:
            py:class:`list` of :py:class:`str`

        :complexity:
            O(1), independent of number of nuts.

        '''
        return sorted(list(self.iter_kinds(codes=codes)))

    def get_deltats(self, kind=None):
        '''
        Get sampling intervals available in selection.

        :param kind:
            If given, get sampling intervals only for selected content type.
        :type kind:
            str

        :complexity:
            O(1), independent of number of nuts.

        :returns: Sorted list of available sampling intervals.
        '''
        return sorted(list(self.iter_deltats(kind=kind)))

    def get_codes(self, kind=None):
        '''
        Get identifier code sequences available in selection.

        :param kind:
            If given, get codes only for selected content type.
        :type kind:
            str

        :complexity:
            O(1), independent of number of nuts.

        :returns: Sorted list of available codes as tuples of strings.
        '''
        return sorted(list(self.iter_codes(kind=kind)))

    def get_counts(self, kind=None):
        '''
        Get number of occurrences of any (kind, codes) combination.

        :param kind:
            If given, get codes only for selected content type.
        :type kind:
            str

        :complexity:
            O(1), independent of number of nuts.

        :returns: ``dict`` with ``counts[kind][codes]`` or ``counts[codes]``
            if kind is not ``None``
        '''
        d = {}
        for kind_id, codes, _, _, count in self._iter_codes_info(kind=kind):
            if kind_id not in d:
                v = d[kind_id] = {}
            else:
                v = d[kind_id]

            if codes not in v:
                v[codes] = 0

            v[codes] += count

        if kind is not None:
            return d[to_kind_id(kind)]
        else:
            return dict((to_kind(kind_id), v) for (kind_id, v) in d.items())

    def glob_codes(self, kind, codes):
        '''
        Find codes matching given patterns.

        :param kind:
            Content kind to be queried.
        :type kind:
            str

        :param codes:
            List of code patterns to query.
        :type codes:
            :py:class:`list` of :py:class:`~pyrockoeost.squirrel.model.Codes`
            objects appropriate for the queried content type, or anything which
            can be converted to such objects.

        :returns:
            List of matches of the form ``[kind_codes_id, codes, deltat]``.
        '''

        kind_id = to_kind_id(kind)
        args = [kind_id]
        pats = codes_patterns_for_kind(kind_id, codes)

        if pats:
            codes_cond = 'AND ( %s ) ' % ' OR '.join(
                    ('kind_codes.codes GLOB ?',) * len(pats))

            args.extend(pat.safe_str for pat in pats)
        else:
            codes_cond = ''

        sql = self._sql('''
            SELECT kind_codes_id, codes, deltat FROM kind_codes
            WHERE
                kind_id == ? ''' + codes_cond)

        return list(map(list, self._conn.execute(sql, args)))

    def update(self, constraint=None, **kwargs):
        '''
        Update or partially update channel and event inventories.

        :param constraint:
            Selection of times or areas to be brought up to date.
        :type constraint:
            :py:class:`~pyrockoeost.squirrel.client.base.Constraint`

        :param \\*\\*kwargs:
            Shortcut for setting ``constraint=Constraint(**kwargs)``.

        This function triggers all attached remote sources, to check for
        updates in the meta-data. The sources will only submit queries when
        their expiration date has passed, or if the selection spans into
        previously unseen times or areas.
        '''

        if constraint is None:
            constraint = client.Constraint(**kwargs)

        for source in self._sources:
            source.update_channel_inventory(self, constraint)
            source.update_event_inventory(self, constraint)

    def update_waveform_promises(self, constraint=None, **kwargs):
        '''
        Permit downloading of remote waveforms.

        :param constraint:
            Remote waveforms compatible with the given constraint are enabled
            for download.
        :type constraint:
            :py:class:`~pyrockoeost.squirrel.client.base.Constraint`

        :param \\*\\*kwargs:
            Shortcut for setting ``constraint=Constraint(**kwargs)``.

        Calling this method permits Squirrel to download waveforms from remote
        sources when processing subsequent waveform requests. This works by
        inserting so called waveform promises into the database. It will look
        into the available channels for each remote source and create a promise
        for each channel compatible with the given constraint. If the promise
        then matches in a waveform request, Squirrel tries to download the
        waveform. If the download is successful, the downloaded waveform is
        added to the Squirrel and the promise is deleted. If the download
        fails, the promise is kept if the reason of failure looks like being
        temporary, e.g. because of a network failure. If the cause of failure
        however seems to be permanent, the promise is deleted so that no
        further attempts are made to download a waveform which might not be
        available from that server at all. To force re-scheduling after a
        permanent failure, call :py:meth:`update_waveform_promises`
        yet another time.
        '''

        if constraint is None:
            constraint = client.Constraint(**kwargs)

        for source in self._sources:
            source.update_waveform_promises(self, constraint)

    def remove_waveform_promises(self, from_database='selection'):
        '''
        Remove waveform promises from live selection or global database.

        Calling this function removes all waveform promises provided by the
        attached sources.

        :param from_database:
            Remove from live selection ``'selection'`` or global database
            ``'global'``.
        '''
        for source in self._sources:
            source.remove_waveform_promises(self, from_database=from_database)

    def update_responses(self, constraint=None, **kwargs):
        if constraint is None:
            constraint = client.Constraint(**kwargs)

        for source in self._sources:
            source.update_response_inventory(self, constraint)

    def get_nfiles(self):
        '''
        Get number of files in selection.
        '''

        sql = self._sql('''SELECT COUNT(*) FROM %(db)s.%(file_states)s''')
        for row in self._conn.execute(sql):
            return row[0]

    def get_nnuts(self):
        '''
        Get number of nuts in selection.
        '''

        sql = self._sql('''SELECT COUNT(*) FROM %(db)s.%(nuts)s''')
        for row in self._conn.execute(sql):
            return row[0]

    def get_total_size(self):
        '''
        Get aggregated file size available in selection.
        '''

        sql = self._sql('''
            SELECT SUM(files.size) FROM %(db)s.%(file_states)s
            INNER JOIN files
                ON %(db)s.%(file_states)s.file_id = files.file_id
        ''')

        for row in self._conn.execute(sql):
            return row[0] or 0

    def get_stats(self):
        '''
        Get statistics on contents available through this selection.
        '''

        kinds = self.get_kinds()
        time_spans = {}
        for kind in kinds:
            time_spans[kind] = self.get_time_span([kind])

        return SquirrelStats(
            nfiles=self.get_nfiles(),
            nnuts=self.get_nnuts(),
            kinds=kinds,
            codes=self.get_codes(),
            total_size=self.get_total_size(),
            counts=self.get_counts(),
            time_spans=time_spans,
            sources=[s.describe() for s in self._sources],
            operators=[op.describe() for op in self._operators])

    @filldocs
    def check(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            ignore=[]):
        '''
        Check for common data/metadata problems.

        %(query_args)s

        :param ignore:
            Problem types to be ignored.
        :type ignore:
            :class:`list` of :class:`str`
            (:py:class:`~pyrockoeost.squirrel.check.SquirrelCheckProblemType`)

        :returns:
            :py:class:`~pyrockoeost.squirrel.check.SquirrelCheck` object
            containing the results of the check.

        See :py:func:`~pyrockoeost.squirrel.check.do_check`.
        '''

        from .check import do_check
        tmin, tmax, codes = self._get_selection_args(
            CHANNEL, obj, tmin, tmax, time, codes)

        return do_check(self, tmin=tmin, tmax=tmax, codes=codes, ignore=ignore)

    def get_content(
            self,
            nut,
            cache_id='default',
            accessor_id='default',
            show_progress=False,
            model='squirrel'):

        '''
        Get and possibly load full content for a given index entry from file.

        Loads the actual content objects (channel, station, waveform, ...) from
        file. For efficiency, sibling content (all stuff in the same file
        segment) will also be loaded as a side effect. The loaded contents are
        cached in the Squirrel object.
        '''

        content_cache = self._content_caches[cache_id]
        if not content_cache.has(nut):

            for nut_loaded in io.iload(
                    nut.file_path,
                    segment=nut.file_segment,
                    format=nut.file_format,
                    database=self._database,
                    update_selection=self,
                    show_progress=show_progress):

                content_cache.put(nut_loaded)

        try:
            return content_cache.get(nut, accessor_id, model)

        except KeyError:
            raise error.NotAvailable(
                'Unable to retrieve content: %s, %s, %s, %s' % nut.key)

    def advance_accessor(self, accessor_id='default', cache_id=None):
        '''
        Notify memory caches about consumer moving to a new data batch.

        :param accessor_id:
            Name of accessing consumer to be advanced.
        :type accessor_id:
            str

        :param cache_id:
            Name of cache to for which the accessor should be advanced. By
            default the named accessor is advanced in all registered caches.
            By default, two caches named ``'default'`` and ``'waveform'`` are
            available.
        :type cache_id:
            str

        See :py:class:`~pyrockoeost.squirrel.cache.ContentCache` for details on how
        Squirrel's memory caching works and can be tuned. Default behaviour is
        to release data when it has not been used in the latest data
        window/batch. If the accessor is never advanced, data is cached
        indefinitely - which is often desired e.g. for station meta-data.
        Methods for consecutive data traversal, like
        :py:meth:`chopper_waveforms` automatically advance and clear
        their accessor.
        '''
        for cache_ in (
                self._content_caches.keys()
                if cache_id is None
                else [cache_id]):

            self._content_caches[cache_].advance_accessor(accessor_id)

    def clear_accessor(self, accessor_id, cache_id=None):
        '''
        Notify memory caches about a consumer having finished.

        :param accessor_id:
            Name of accessor to be cleared.
        :type accessor_id:
            str

        :param cache_id:
            Name of cache for which the accessor should be cleared. By default
            the named accessor is cleared from all registered caches. By
            default, two caches named ``'default'`` and ``'waveform'`` are
            available.
        :type cache_id:
            str

        Calling this method clears all references to cache entries held by the
        named accessor. Cache entries are then freed if not referenced by any
        other accessor.
        '''

        for cache_ in (
                self._content_caches.keys()
                if cache_id is None
                else [cache_id]):

            self._content_caches[cache_].clear_accessor(accessor_id)

    def get_cache_stats(self, cache_id):
        return self._content_caches[cache_id].get_stats()

    @filldocs
    def get_stations(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            model='squirrel'):

        '''
        Get stations matching given constraints.

        %(query_args)s

        :param model:
            Select object model for returned values: ``'squirrel'`` to get
            Squirrel station objects or ``'pyrockoeost'`` to get pyrockoeost station
            objects with channel information attached.
        :type model:
            str

        :returns:
            List of :py:class:`pyrockoeost.squirrel.Station
            <pyrockoeost.squirrel.model.Station>` objects by default or list of
            :py:class:`pyrockoeost.model.Station <pyrockoeost.model.station.Station>`
            objects if ``model='pyrockoeost'`` is requested.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        if model == 'pyrockoeost':
            return self._get_pyrockoeost_stations(obj, tmin, tmax, time, codes)
        elif model in ('squirrel', 'stationxml', 'stationxml+'):
            args = self._get_selection_args(
                STATION, obj, tmin, tmax, time, codes)

            nuts = sorted(
                self.iter_nuts('station', *args), key=lambda nut: nut.dkey)

            return [self.get_content(nut, model=model) for nut in nuts]
        else:
            raise ValueError('Invalid station model: %s' % model)

    @filldocs
    def get_channels(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            model='squirrel'):

        '''
        Get channels matching given constraints.

        %(query_args)s

        :returns:
            List of :py:class:`~pyrockoeost.squirrel.model.Channel` objects.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        args = self._get_selection_args(
            CHANNEL, obj, tmin, tmax, time, codes)

        nuts = sorted(
            self.iter_nuts('channel', *args), key=lambda nut: nut.dkey)

        return [self.get_content(nut, model=model) for nut in nuts]

    @filldocs
    def get_sensors(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None):

        '''
        Get sensors matching given constraints.

        %(query_args)s

        :returns:
            List of :py:class:`~pyrockoeost.squirrel.model.Sensor` objects.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        tmin, tmax, codes = self._get_selection_args(
            CHANNEL, obj, tmin, tmax, time, codes)

        if codes is not None:
            codes = codes_patterns_list(
                (entry.replace(channel=entry.channel[:-1] + '?')
                 if entry.channel != '*' else entry)
                for entry in codes)

        nuts = sorted(
            self.iter_nuts(
                'channel', tmin, tmax, codes), key=lambda nut: nut.dkey)

        return [
            sensor for sensor in model.Sensor.from_channels(
                self.get_content(nut) for nut in nuts)
            if match_time_span(tmin, tmax, sensor)]

    @filldocs
    def get_responses(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            model='squirrel'):

        '''
        Get instrument responses matching given constraints.

        %(query_args)s

        :param model:
            Select data model for returned objects. Choices: ``'squirrel'``,
            ``'stationxml'``, ``'stationxml+'``. See return value description.
        :type model:
            str

        :returns:
            List of :py:class:`~pyrockoeost.squirrel.model.Response` if ``model ==
            'squirrel'`` or list of :py:class:`~pyrockoeost.io.fdsn.FDSNStationXML`
            if ``model == 'stationxml'`` or list of
            (:py:class:`~pyrockoeost.squirrel.model.Response`,
            :py:class:`~pyrockoeost.io.fdsn.FDSNStationXML`) if ``model ==
            'stationxml+'``.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        args = self._get_selection_args(
            RESPONSE, obj, tmin, tmax, time, codes)

        nuts = sorted(
            self.iter_nuts('response', *args), key=lambda nut: nut.dkey)

        return [self.get_content(nut, model=model) for nut in nuts]

    @filldocs
    def get_response(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            model='squirrel', on_duplicate='raise'):

        '''
        Get instrument response matching given constraints.

        %(query_args)s

        :param model:
            Select data model for returned object. Choices: ``'squirrel'``,
            ``'stationxml'``, ``'stationxml+'``. See return value description.
        :type model:
            str

        :param on_duplicate:
            Determines how duplicates/multiple matching responses are handled.
            Choices: ``'raise'`` - raise
            :py:exc:`~pyrockoeost.squirrel.error.Duplicate`, ``'warn'`` - emit a
            warning and return first match, ``'ignore'`` - silently return
            first match.
        :type on_duplicate:
            str

        :returns:
            :py:class:`~pyrockoeost.squirrel.model.Response` if
            ``model == 'squirrel'`` or
            :py:class:`~pyrockoeost.io.fdsn.FDSNStationXML` if ``model ==
            'stationxml'`` or
            (:py:class:`~pyrockoeost.squirrel.model.Response`,
            :py:class:`~pyrockoeost.io.fdsn.FDSNStationXML`) if ``model ==
            'stationxml+'``.

        Same as :py:meth:`get_responses` but returning exactly one response.
        Raises :py:exc:`~pyrockoeost.squirrel.error.NotAvailable` if none is
        available. Duplicates are handled according to the ``on_duplicate``
        argument.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        if model == 'stationxml':
            model_ = 'stationxml+'
        else:
            model_ = model

        responses = self.get_responses(
            obj, tmin, tmax, time, codes, model=model_)
        if len(responses) == 0:
            raise error.NotAvailable(
                'No instrument response available (%s).'
                % self._get_selection_args_str(
                    RESPONSE, obj, tmin, tmax, time, codes))

        elif len(responses) > 1:

            if on_duplicate in ('raise', 'warn'):
                if model_ == 'squirrel':
                    resps_sq = responses
                elif model_ == 'stationxml+':
                    resps_sq = [resp[0] for resp in responses]
                else:
                    raise ValueError('Invalid response model: %s' % model)

                rinfo = ':\n' + '\n'.join(
                    '  ' + resp.summary for resp in resps_sq)

                message = \
                    'Multiple instrument responses matching given ' \
                    'constraints (%s)%s%s' % (
                        self._get_selection_args_str(
                            RESPONSE, obj, tmin, tmax, time, codes),
                        ' -> using first' if on_duplicate == 'warn' else '',
                        rinfo)

            if on_duplicate == 'raise':
                raise error.Duplicate(message)

            elif on_duplicate == 'warn':
                logger.warning(message)

            elif on_duplicate == 'ignore':
                pass

            else:
                ValueError(
                    'Invalid argument for on_duplicate: %s' % on_duplicate)

        if model == 'stationxml':
            return responses[0][1]
        else:
            return responses[0]

    @filldocs
    def get_events(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None):

        '''
        Get events matching given constraints.

        %(query_args)s

        :returns:
            List of :py:class:`~pyrockoeost.model.event.Event` objects.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        args = self._get_selection_args(EVENT, obj, tmin, tmax, time, codes)
        nuts = sorted(
            self.iter_nuts('event', *args), key=lambda nut: nut.dkey)

        return [self.get_content(nut) for nut in nuts]

    def _redeem_promises(self, *args, codes_exclude=None, order_only=False):

        def split_promise(order):
            self._split_nuts(
                'waveform_promise',
                order.tmin, order.tmax,
                codes=order.codes,
                path=order.source_id)

        tmin, tmax, _ = args

        waveforms = list(self.iter_nuts('waveform', *args))
        promises = list(self.iter_nuts('waveform_promise', *args))
        if codes_exclude is not None:
            promises = [
                promise for promise in promises
                if promise.codes not in codes_exclude]

        codes_to_avail = defaultdict(list)
        for nut in waveforms:
            codes_to_avail[nut.codes].append((nut.tmin, nut.tmax))

        def tts(x):
            if isinstance(x, tuple):
                return tuple(tts(e) for e in x)
            elif isinstance(x, list):
                return list(tts(e) for e in x)
            else:
                return util.time_to_str(x)

        orders = []
        for promise in promises:
            waveforms_avail = codes_to_avail[promise.codes]
            for block_tmin, block_tmax in blocks(
                    max(tmin, promise.tmin),
                    min(tmax, promise.tmax),
                    promise.deltat):

                orders.append(
                    WaveformOrder(
                        source_id=promise.file_path,
                        codes=promise.codes,
                        tmin=block_tmin,
                        tmax=block_tmax,
                        deltat=promise.deltat,
                        gaps=gaps(waveforms_avail, block_tmin, block_tmax)))

        orders_noop, orders = lpick(lambda order: order.gaps, orders)

        order_keys_noop = set(order_key(order) for order in orders_noop)
        if len(order_keys_noop) != 0 or len(orders_noop) != 0:
            logger.info(
                'Waveform orders already satisified with cached/local data: '
                '%i (%i)' % (len(order_keys_noop), len(orders_noop)))

        for order in orders_noop:
            split_promise(order)

        if order_only:
            if orders:
                self._pending_orders.extend(orders)
                logger.info(
                    'Enqueuing %i waveform order%s.'
                    % len_plural(orders))
            return
        else:
            if self._pending_orders:
                orders.extend(self._pending_orders)
                logger.info(
                    'Adding %i previously enqueued order%s.'
                    % len_plural(self._pending_orders))

            self._pending_orders = []

        source_ids = []
        sources = {}
        for source in self._sources:
            if isinstance(source, fdsn.FDSNSource):
                source_ids.append(source._source_id)
                sources[source._source_id] = source

        source_priority = dict(
            (source_id, i) for (i, source_id) in enumerate(source_ids))

        order_groups = defaultdict(list)
        for order in orders:
            order_groups[order_key(order)].append(order)

        for k, order_group in order_groups.items():
            order_group.sort(
                key=lambda order: source_priority[order.source_id])

        n_order_groups = len(order_groups)

        if len(order_groups) != 0 or len(orders) != 0:
            logger.info(
                'Waveform orders standing for download: %i (%i)'
                % (len(order_groups), len(orders)))

            task = make_task('Waveform orders processed', n_order_groups)
        else:
            task = None

        def release_order_group(order):
            okey = order_key(order)
            for followup in order_groups[okey]:
                split_promise(followup)

            del order_groups[okey]

            if task:
                task.update(n_order_groups - len(order_groups))

        def noop(order):
            pass

        def success(order):
            release_order_group(order)
            split_promise(order)

        def batch_add(paths):
            self.add(paths)

        calls = queue.Queue()

        def enqueue(f):
            def wrapper(*args):
                calls.put((f, args))

            return wrapper

        while order_groups:

            orders_now = []
            empty = []
            for k, order_group in order_groups.items():
                try:
                    orders_now.append(order_group.pop(0))
                except IndexError:
                    empty.append(k)

            for k in empty:
                del order_groups[k]

            by_source_id = defaultdict(list)
            for order in orders_now:
                by_source_id[order.source_id].append(order)

            threads = []
            for source_id in by_source_id:
                def download():
                    try:
                        sources[source_id].download_waveforms(
                            by_source_id[source_id],
                            success=enqueue(success),
                            error_permanent=enqueue(split_promise),
                            error_temporary=noop,
                            batch_add=enqueue(batch_add))

                    finally:
                        calls.put(None)

                thread = threading.Thread(target=download)
                thread.start()
                threads.append(thread)

            ndone = 0
            while ndone < len(threads):
                ret = calls.get()
                if ret is None:
                    ndone += 1
                else:
                    ret[0](*ret[1])

            for thread in threads:
                thread.join()

            if task:
                task.update(n_order_groups - len(order_groups))

        if task:
            task.done()

    @filldocs
    def get_waveform_nuts(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            codes_exclude=None, order_only=False):

        '''
        Get waveform content entities matching given constraints.

        %(query_args)s

        Like :py:meth:`get_nuts` with ``kind='waveform'`` but additionally
        resolves matching waveform promises (downloads waveforms from remote
        sources).

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        args = self._get_selection_args(WAVEFORM, obj, tmin, tmax, time, codes)
        self._redeem_promises(
            *args, codes_exclude=codes_exclude, order_only=order_only)
        nuts = sorted(
            self.iter_nuts('waveform', *args), key=lambda nut: nut.dkey)

        if codes_exclude is not None:
            nuts = [nut for nut in nuts if nut.codes not in codes_exclude]

        return nuts

    @filldocs
    def have_waveforms(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None):

        '''
        Check if any waveforms or waveform promises are available for given
        constraints.

        %(query_args)s
        '''

        args = self._get_selection_args(WAVEFORM, obj, tmin, tmax, time, codes)
        return bool(list(
                self.iter_nuts('waveform', *args, limit=1))) \
            or bool(list(
                self.iter_nuts('waveform_promise', *args, limit=1)))

    @filldocs
    def get_waveforms(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            codes_exclude=None, uncut=False, want_incomplete=True, degap=True,
            maxgap=5, maxlap=None, snap=None, include_last=False,
            load_data=True, accessor_id='default', operator_params=None,
            order_only=False, channel_priorities=None, target_deltat=None):

        '''
        Get waveforms matching given constraints.

        %(query_args)s

        :param uncut:
            Set to ``True``, to disable cutting traces to [``tmin``, ``tmax``]
            and to disable degapping/deoverlapping. Returns untouched traces as
            they are read from file segment. File segments are always read in
            their entirety.
        :type uncut:
            bool

        :param want_incomplete:
            If ``True``, gappy/incomplete traces are included in the result.
        :type want_incomplete:
            bool

        :param degap:
            If ``True``, connect traces and remove gaps and overlaps.
        :type degap:
            bool

        :param maxgap:
            Maximum gap size in samples which is filled with interpolated
            samples when ``degap`` is ``True``.
        :type maxgap:
            int

        :param maxlap:
            Maximum overlap size in samples which is removed when ``degap`` is
            ``True``.
        :type maxlap:
            int

        :param snap:
            Rounding functions used when computing sample index from time
            instance, for trace start and trace end, respectively. By default,
            ``(round, round)`` is used.
        :type snap:
             tuple of 2 callables

        :param include_last:
            If ``True``, add one more sample to the returned traces (the sample
            which would be the first sample of a query with ``tmin`` set to the
            current value of ``tmax``).
        :type include_last:
            bool

        :param load_data:
            If ``True``, waveform data samples are read from files (or cache).
            If ``False``, meta-information-only traces are returned (dummy
            traces with no data samples).
        :type load_data:
            bool

        :param accessor_id:
            Name of consumer on who's behalf data is accessed. Used in cache
            management (see :py:mod:`~pyrockoeost.squirrel.cache`). Used as a key
            to distinguish different points of extraction for the decision of
            when to release cached waveform data. Should be used when data is
            alternately extracted from more than one region / selection.
        :type accessor_id:
            str

        See :py:meth:`iter_nuts` for details on time span matching.

        Loaded data is kept in memory (at least) until
        :py:meth:`clear_accessor` has been called or
        :py:meth:`advance_accessor` has been called two consecutive times
        without data being accessed between the two calls (by this accessor).
        Data may still be further kept in the memory cache if held alive by
        consumers with a different ``accessor_id``.
        '''

        tmin, tmax, codes = self._get_selection_args(
            WAVEFORM, obj, tmin, tmax, time, codes)

        if channel_priorities is not None:
            return self._get_waveforms_prioritized(
                tmin=tmin, tmax=tmax, codes=codes,
                uncut=uncut, want_incomplete=want_incomplete, degap=degap,
                maxgap=maxgap, maxlap=maxlap, snap=snap,
                include_last=include_last, load_data=load_data,
                accessor_id=accessor_id, operator_params=operator_params,
                order_only=order_only, channel_priorities=channel_priorities,
                target_deltat=target_deltat)

        self_tmin, self_tmax = self.get_time_span(
            ['waveform', 'waveform_promise'])

        if None in (self_tmin, self_tmax):
            logger.warning(
                'No waveforms available.')
            return []

        tmin = tmin if tmin is not None else self_tmin
        tmax = tmax if tmax is not None else self_tmax

        if codes is not None and len(codes) == 1:
            # TODO: fix for multiple / mixed codes
            operator = self.get_operator(codes[0])
            if operator is not None:
                return operator.get_waveforms(
                    self, codes[0],
                    tmin=tmin, tmax=tmax,
                    uncut=uncut, want_incomplete=want_incomplete, degap=degap,
                    maxgap=maxgap, maxlap=maxlap, snap=snap,
                    include_last=include_last, load_data=load_data,
                    accessor_id=accessor_id, params=operator_params)

        nuts = self.get_waveform_nuts(
            obj, tmin, tmax, time, codes, codes_exclude=codes_exclude,
            order_only=order_only)

        if order_only:
            return []

        if load_data:
            traces = [
                self.get_content(nut, 'waveform', accessor_id) for nut in nuts]

        else:
            traces = [
                trace.Trace(**nut.trace_kwargs) for nut in nuts]

        if uncut:
            return traces

        if snap is None:
            snap = (round, round)

        chopped = []
        for tr in traces:
            if not load_data and tr.ydata is not None:
                tr = tr.copy(data=False)
                tr.ydata = None

            try:
                chopped.append(tr.chop(
                    tmin, tmax,
                    inplace=False,
                    snap=snap,
                    include_last=include_last))

            except trace.NoData:
                pass

        processed = self._process_chopped(
            chopped, degap, maxgap, maxlap, want_incomplete, tmin, tmax)

        return processed

    def _get_waveforms_prioritized(
            self, tmin=None, tmax=None, codes=None,
            channel_priorities=None, target_deltat=None, **kwargs):

        trs_all = []
        codes_have = set()
        for channel in channel_priorities:
            assert len(channel) == 2
            if codes is not None:
                codes_now = [
                    codes_.replace(channel=channel+'?') for codes_ in codes]
            else:
                codes_now = model.CodesNSLCE('*', '*', '*', channel+'?')

            codes_exclude_now = set(
                codes_.replace(channel=channel+codes_.channel[-1])
                for codes_ in codes_have)

            trs = self.get_waveforms(
                tmin=tmin,
                tmax=tmax,
                codes=codes_now,
                codes_exclude=codes_exclude_now,
                **kwargs)

            codes_have.update(set(tr.codes for tr in trs))
            trs_all.extend(trs)

        return trs_all

    @filldocs
    def chopper_waveforms(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            tinc=None, tpad=0.,
            want_incomplete=True, snap_window=False,
            degap=True, maxgap=5, maxlap=None,
            snap=None, include_last=False, load_data=True,
            accessor_id=None, clear_accessor=True, operator_params=None,
            grouping=None, channel_priorities=None, target_deltat=None):

        '''
        Iterate window-wise over waveform archive.

        %(query_args)s

        :param tinc:
            Time increment (window shift time) (default uses ``tmax-tmin``).
        :type tinc:
            timestamp

        :param tpad:
            Padding time appended on either side of the data window (window
            overlap is ``2*tpad``).
        :type tpad:
            timestamp

        :param want_incomplete:
            If ``True``, gappy/incomplete traces are included in the result.
        :type want_incomplete:
            bool

        :param snap_window:
            If ``True``, start time windows at multiples of tinc with respect
            to system time zero.
        :type snap_window:
            bool

        :param degap:
            If ``True``, connect traces and remove gaps and overlaps.
        :type degap:
            bool

        :param maxgap:
            Maximum gap size in samples which is filled with interpolated
            samples when ``degap`` is ``True``.
        :type maxgap:
            int

        :param maxlap:
            Maximum overlap size in samples which is removed when ``degap`` is
            ``True``.
        :type maxlap:
            int

        :param snap:
            Rounding functions used when computing sample index from time
            instance, for trace start and trace end, respectively. By default,
            ``(round, round)`` is used.
        :type snap:
             tuple of 2 callables

        :param include_last:
            If ``True``, add one more sample to the returned traces (the sample
            which would be the first sample of a query with ``tmin`` set to the
            current value of ``tmax``).
        :type include_last:
            bool

        :param load_data:
            If ``True``, waveform data samples are read from files (or cache).
            If ``False``, meta-information-only traces are returned (dummy
            traces with no data samples).
        :type load_data:
            bool

        :param accessor_id:
            Name of consumer on who's behalf data is accessed. Used in cache
            management (see :py:mod:`~pyrockoeost.squirrel.cache`). Used as a key
            to distinguish different points of extraction for the decision of
            when to release cached waveform data. Should be used when data is
            alternately extracted from more than one region / selection.
        :type accessor_id:
            str

        :param clear_accessor:
            If ``True`` (default), :py:meth:`clear_accessor` is called when the
            chopper finishes. Set to ``False`` to keep loaded waveforms in
            memory when the generator returns.
        :type clear_accessor:
            bool

        :param grouping:
            By default, traversal over the data is over time and all matching
            traces of a time window are yielded. Using this option, it is
            possible to traverse the data first by group (e.g. station or
            network) and second by time. This can reduce the number of traces
            in each batch and thus reduce the memory footprint of the process.
        :type grouping:
            :py:class:`~pyrockoeost.squirrel.operator.Grouping`

        :yields:
            A list of :py:class:`~pyrockoeost.trace.Trace` objects for every
            extracted time window.

        See :py:meth:`iter_nuts` for details on time span matching.
        '''

        tmin, tmax, codes = self._get_selection_args(
            WAVEFORM, obj, tmin, tmax, time, codes)

        self_tmin, self_tmax = self.get_time_span(
            ['waveform', 'waveform_promise'])

        if None in (self_tmin, self_tmax):
            logger.warning(
                'Content has undefined time span. No waveforms and no '
                'waveform promises?')
            return

        if snap_window and tinc is not None:
            tmin = tmin if tmin is not None else self_tmin
            tmax = tmax if tmax is not None else self_tmax
            tmin = math.floor(tmin / tinc) * tinc
            tmax = math.ceil(tmax / tinc) * tinc
        else:
            tmin = tmin if tmin is not None else self_tmin + tpad
            tmax = tmax if tmax is not None else self_tmax - tpad

        tinc = tinc if tinc is not None else tmax - tmin

        try:
            if accessor_id is None:
                accessor_id = 'chopper%i' % self._n_choppers_active

            self._n_choppers_active += 1

            eps = tinc * 1e-6
            if tinc != 0.0:
                nwin = int(((tmax - eps) - tmin) / tinc) + 1
            else:
                nwin = 1

            if grouping is None:
                codes_list = [codes]
            else:
                operator = Operator(
                    filtering=CodesPatternFiltering(codes=codes),
                    grouping=grouping)

                available = set(self.get_codes(kind='waveform'))
                available.update(self.get_codes(kind='waveform_promise'))
                operator.update_mappings(sorted(available))

                codes_list = [
                    codes_patterns_list(scl)
                    for scl in operator.iter_in_codes()]

            ngroups = len(codes_list)
            for igroup, scl in enumerate(codes_list):
                for iwin in range(nwin):
                    wmin, wmax = tmin+iwin*tinc, min(tmin+(iwin+1)*tinc, tmax)

                    chopped = self.get_waveforms(
                        tmin=wmin-tpad,
                        tmax=wmax+tpad,
                        codes=scl,
                        snap=snap,
                        include_last=include_last,
                        load_data=load_data,
                        want_incomplete=want_incomplete,
                        degap=degap,
                        maxgap=maxgap,
                        maxlap=maxlap,
                        accessor_id=accessor_id,
                        operator_params=operator_params,
                        channel_priorities=channel_priorities,
                        target_deltat=target_deltat)

                    self.advance_accessor(accessor_id)

                    yield Batch(
                        tmin=wmin,
                        tmax=wmax,
                        i=iwin,
                        n=nwin,
                        igroup=igroup,
                        ngroups=ngroups,
                        traces=chopped)

        finally:
            self._n_choppers_active -= 1
            if clear_accessor:
                self.clear_accessor(accessor_id, 'waveform')

    def _process_chopped(
            self, chopped, degap, maxgap, maxlap, want_incomplete, tmin, tmax):

        chopped.sort(key=lambda a: a.full_id)
        if degap:
            chopped = trace.degapper(chopped, maxgap=maxgap, maxlap=maxlap)

        if not want_incomplete:
            chopped_weeded = []
            for tr in chopped:
                emin = tr.tmin - tmin
                emax = tr.tmax + tr.deltat - tmax
                if (abs(emin) <= 0.5*tr.deltat and abs(emax) <= 0.5*tr.deltat):
                    chopped_weeded.append(tr)

                elif degap:
                    if (0. < emin <= 5. * tr.deltat
                            and -5. * tr.deltat <= emax < 0.):

                        tr.extend(tmin, tmax-tr.deltat, fillmethod='repeat')
                        chopped_weeded.append(tr)

            chopped = chopped_weeded

        return chopped

    def _get_pyrockoeost_stations(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None):

        from pyrockoeost import model as pmodel

        if codes is not None:
            codes = codes_patterns_for_kind(STATION, codes)

        by_nsl = defaultdict(lambda: (list(), list()))
        for station in self.get_stations(obj, tmin, tmax, time, codes):
            sargs = station._get_pyrockoeost_station_args()
            by_nsl[station.codes.nsl][0].append(sargs)

        if codes is not None:
            codes = [model.CodesNSLCE(c) for c in codes]

        for channel in self.get_channels(obj, tmin, tmax, time, codes):
            sargs = channel._get_pyrockoeost_station_args()
            sargs_list, channels_list = by_nsl[channel.codes.nsl]
            sargs_list.append(sargs)
            channels_list.append(channel)

        pstations = []
        nsls = list(by_nsl.keys())
        nsls.sort()
        for nsl in nsls:
            sargs_list, channels_list = by_nsl[nsl]
            sargs = util.consistency_merge(
                [('',) + x for x in sargs_list])

            by_c = defaultdict(list)
            for ch in channels_list:
                by_c[ch.codes.channel].append(ch._get_pyrockoeost_channel_args())

            chas = list(by_c.keys())
            chas.sort()
            pchannels = []
            for cha in chas:
                list_of_cargs = by_c[cha]
                cargs = util.consistency_merge(
                    [('',) + x for x in list_of_cargs])
                pchannels.append(pmodel.Channel(*cargs))

            pstations.append(
                pmodel.Station(*sargs, channels=pchannels))

        return pstations

    @property
    def pile(self):

        '''
        Emulates the older :py:class:`pyrockoeost.pile.Pile` interface.

        This property exposes a :py:class:`pyrockoeost.squirrel.pile.Pile` object,
        which emulates most of the older :py:class:`pyrockoeost.pile.Pile` methods
        but uses the fluffy power of the Squirrel under the hood.

        This interface can be used as a drop-in replacement for piles which are
        used in existing scripts and programs for efficient waveform data
        access. The Squirrel-based pile scales better for large datasets. Newer
        scripts should use Squirrel's native methods to avoid the emulation
        overhead.
        '''
        from . import pile

        if self._pile is None:
            self._pile = pile.Pile(self)

        return self._pile

    def snuffle(self):
        '''
        Look at dataset in Snuffler.
        '''
        self.pile.snuffle()

    def _gather_codes_keys(self, kind, gather, selector):
        return set(
            gather(codes)
            for codes in self.iter_codes(kind)
            if selector is None or selector(codes))

    def __str__(self):
        return str(self.get_stats())

    def get_coverage(
            self, kind, tmin=None, tmax=None, codes=None, limit=None):

        '''
        Get coverage information.

        Get information about strips of gapless data coverage.

        :param kind:
            Content kind to be queried.
        :type kind:
            str

        :param tmin:
            Start time of query interval.
        :type tmin:
            timestamp

        :param tmax:
            End time of query interval.
        :type tmax:
            timestamp

        :param codes:
            If given, restrict query to given content codes patterns.
        :type codes:
            :py:class:`list` of :py:class:`~pyrockoeost.squirrel.model.Codes`
            objects appropriate for the queried content type, or anything which
            can be converted to such objects.

        :param limit:
            Limit query to return only up to a given maximum number of entries
            per matching time series (without setting this option, very gappy
            data could cause the query to execute for a very long time).
        :type limit:
            int

        :returns:
            Information about time spans covered by the requested time series
            data.
        :rtype:
            :py:class:`list` of :py:class:`Coverage` objects
        '''

        tmin_seconds, tmin_offset = model.tsplit(tmin)
        tmax_seconds, tmax_offset = model.tsplit(tmax)
        kind_id = to_kind_id(kind)

        codes_info = list(self._iter_codes_info(kind=kind))

        kdata_all = []
        if codes is None:
            for _, codes_entry, deltat, kind_codes_id, _ in codes_info:
                kdata_all.append(
                    (codes_entry, kind_codes_id, codes_entry, deltat))

        else:
            for codes_entry in codes:
                pattern = to_codes(kind_id, codes_entry)
                for _, codes_entry, deltat, kind_codes_id, _ in codes_info:
                    if model.match_codes(pattern, codes_entry):
                        kdata_all.append(
                            (pattern, kind_codes_id, codes_entry, deltat))

        kind_codes_ids = [x[1] for x in kdata_all]

        counts_at_tmin = {}
        if tmin is not None:
            for nut in self.iter_nuts(
                    kind, tmin, tmin, kind_codes_ids=kind_codes_ids):

                k = nut.codes, nut.deltat
                if k not in counts_at_tmin:
                    counts_at_tmin[k] = 0

                counts_at_tmin[k] += 1

        coverages = []
        for pattern, kind_codes_id, codes_entry, deltat in kdata_all:
            entry = [pattern, codes_entry, deltat, None, None, []]
            for i, order in [(0, 'ASC'), (1, 'DESC')]:
                sql = self._sql('''
                    SELECT
                        time_seconds,
                        time_offset
                    FROM %(db)s.%(coverage)s
                    WHERE
                        kind_codes_id == ?
                    ORDER BY
                        kind_codes_id ''' + order + ''',
                        time_seconds ''' + order + ''',
                        time_offset ''' + order + '''
                    LIMIT 1
                ''')

                for row in self._conn.execute(sql, [kind_codes_id]):
                    entry[3+i] = model.tjoin(row[0], row[1])

            if None in entry[3:5]:
                continue

            args = [kind_codes_id]

            sql_time = ''
            if tmin is not None:
                # intentionally < because (== tmin) is queried from nuts
                sql_time += ' AND ( ? < time_seconds ' \
                    'OR ( ? == time_seconds AND ? < time_offset ) ) '
                args.extend([tmin_seconds, tmin_seconds, tmin_offset])

            if tmax is not None:
                sql_time += ' AND ( time_seconds < ? ' \
                    'OR ( ? == time_seconds AND time_offset <= ? ) ) '
                args.extend([tmax_seconds, tmax_seconds, tmax_offset])

            sql_limit = ''
            if limit is not None:
                sql_limit = ' LIMIT ?'
                args.append(limit)

            sql = self._sql('''
                SELECT
                    time_seconds,
                    time_offset,
                    step
                FROM %(db)s.%(coverage)s
                WHERE
                    kind_codes_id == ?
                    ''' + sql_time + '''
                ORDER BY
                    kind_codes_id,
                    time_seconds,
                    time_offset
            ''' + sql_limit)

            rows = list(self._conn.execute(sql, args))

            if limit is not None and len(rows) == limit:
                entry[-1] = None
            else:
                counts = counts_at_tmin.get((codes_entry, deltat), 0)
                tlast = None
                if tmin is not None:
                    entry[-1].append((tmin, counts))
                    tlast = tmin

                for row in rows:
                    t = model.tjoin(row[0], row[1])
                    counts += row[2]
                    entry[-1].append((t, counts))
                    tlast = t

                if tmax is not None and (tlast is None or tlast != tmax):
                    entry[-1].append((tmax, counts))

            coverages.append(model.Coverage.from_values(entry + [kind_id]))

        return coverages

    def get_stationxml(
            self, obj=None, tmin=None, tmax=None, time=None, codes=None,
            level='response', on_error='raise'):

        '''
        Get station/channel/response metadata in StationXML representation.

        %(query_args)s

        :returns:
            :py:class:`~pyrockoeost.io.stationxml.FDSNStationXML` object.
        '''

        if level not in ('network', 'station', 'channel', 'response'):
            raise ValueError('Invalid level: %s' % level)

        tmin, tmax, codes = self._get_selection_args(
            CHANNEL, obj, tmin, tmax, time, codes)

        def tts(t):
            if t is None:
                return '<none>'
            else:
                return util.tts(t, format='%Y-%m-%d %H:%M:%S')

        if on_error == 'ignore':
            def handle_error(exc):
                pass

        elif on_error == 'warn':
            def handle_error(exc):
                logger.warning(str(exc))

        elif on_error == 'raise':
            def handle_error(exc):
                raise exc

        def use_first(node_type_name, codes, k, group):
            if on_error == 'warn':
                logger.warning(
                    'Duplicates for %s %s, %s - %s -> using first' % (
                        node_type_name,
                        '.'.join(codes),
                        tts(k[0]), tts(k[1])))

            return group[0]

        def deduplicate(node_type_name, codes, nodes):
            groups = defaultdict(list)
            for node in nodes:
                k = (node.start_date, node.end_date)
                groups[k].append(node)

            return [
                use_first(node_type_name, codes, k, group)
                for (k, group) in groups.items()]

        filtering = CodesPatternFiltering(codes=codes)

        nslcs = list(set(
            codes.nslc for codes in
            filtering.filter(self.get_codes(kind='channel'))))

        from pyrockoeost.io import stationxml as sx

        networks = []
        for net, stas in prefix_tree(nslcs):
            network = sx.Network(code=net)
            networks.append(network)

            if level not in ('station', 'channel', 'response'):
                continue

            for sta, locs in stas:
                stations = self.get_stations(
                    tmin=tmin,
                    tmax=tmax,
                    codes=(net, sta, '*'),
                    model='stationxml')

                if on_error != 'raise':
                    stations = deduplicate(
                        'Station', (net, sta), stations)

                errors = sx.check_overlaps(
                    'Station', (net, sta), stations)

                if errors:
                    handle_error(error.Duplicate(
                        'Overlapping/duplicate station info:\n  %s'
                        % '\n  '.join(errors)))

                network.station_list.extend(stations)

                if level not in ('channel', 'response'):
                    continue

                for loc, chas in locs:
                    for cha, _ in chas:
                        channels = self.get_channels(
                            tmin=tmin,
                            tmax=tmax,
                            codes=(net, sta, loc, cha),
                            model='stationxml')

                        if on_error != 'raise':
                            channels = deduplicate(
                                'Channel', (net, sta, loc, cha), channels)

                        errors = sx.check_overlaps(
                            'Channel', (net, sta, loc, cha), channels)

                        if errors:
                            handle_error(error.Duplicate(
                                'Overlapping/duplicate channel info:\n  %s'
                                % '\n  '.join(errors)))

                        for channel in channels:
                            station = sx.find_containing(stations, channel)
                            if station is not None:
                                station.channel_list.append(channel)
                            else:
                                handle_error(error.NotAvailable(
                                    'No station or station epoch found '
                                    'for channel: %s' % '.'.join(
                                        (net, sta, loc, cha))))

                                continue

                            if level != 'response':
                                continue

                            try:
                                response_sq, response_sx = self.get_response(
                                    codes=(net, sta, loc, cha),
                                    tmin=channel.start_date,
                                    tmax=channel.end_date,
                                    model='stationxml+',
                                    on_duplicate=on_error)

                            except error.NotAvailable as e:
                                handle_error(e)
                                continue

                            if not (
                                    sx.eq_open(
                                        channel.start_date, response_sq.tmin)
                                    and sx.eq_open(
                                        channel.end_date, response_sq.tmax)):

                                handle_error(error.Inconsistencies(
                                    'Response time span does not match '
                                    'channel time span: %s' % '.'.join(
                                        (net, sta, loc, cha))))

                            channel.response = response_sx

        return sx.FDSNStationXML(
            source='Generated by pyrockoeost Squirrel.',
            network_list=networks)

    def add_operator(self, op):
        self._operators.append(op)

    def update_operator_mappings(self):
        available = self.get_codes(kind=('channel'))

        for operator in self._operators:
            operator.update_mappings(available, self._operator_registry)

    def iter_operator_mappings(self):
        for operator in self._operators:
            for in_codes, out_codes in operator.iter_mappings():
                yield operator, in_codes, out_codes

    def get_operator_mappings(self):
        return list(self.iter_operator_mappings())

    def get_operator(self, codes):
        try:
            return self._operator_registry[codes][0]
        except KeyError:
            return None

    def get_operator_group(self, codes):
        try:
            return self._operator_registry[codes]
        except KeyError:
            return None, (None, None, None)

    def iter_operator_codes(self):
        for _, _, out_codes in self.iter_operator_mappings():
            for codes in out_codes:
                yield codes

    def get_operator_codes(self):
        return list(self.iter_operator_codes())

    def print_tables(self, table_names=None, stream=None):
        '''
        Dump raw database tables in textual form (for debugging purposes).

        :param table_names:
            Names of tables to be dumped or ``None`` to dump all.
        :type table_names:
            :py:class:`list` of :py:class:`str`

        :param stream:
            Open file or ``None`` to dump to standard output.
        '''

        if stream is None:
            stream = sys.stdout

        if isinstance(table_names, str):
            table_names = [table_names]

        if table_names is None:
            table_names = [
                'selection_file_states',
                'selection_nuts',
                'selection_kind_codes_count',
                'files', 'nuts', 'kind_codes', 'kind_codes_count']

        m = {
            'selection_file_states': '%(db)s.%(file_states)s',
            'selection_nuts': '%(db)s.%(nuts)s',
            'selection_kind_codes_count': '%(db)s.%(kind_codes_count)s',
            'files': 'files',
            'nuts': 'nuts',
            'kind_codes': 'kind_codes',
            'kind_codes_count': 'kind_codes_count'}

        for table_name in table_names:
            self._database.print_table(
                m[table_name] % self._names, stream=stream)


class SquirrelStats(Object):
    '''
    Container to hold statistics about contents available from a Squirrel.

    See also :py:meth:`Squirrel.get_stats`.
    '''

    nfiles = Int.T(
        help='Number of files in selection.')
    nnuts = Int.T(
        help='Number of index nuts in selection.')
    codes = List.T(
        Tuple.T(content_t=String.T()),
        help='Available code sequences in selection, e.g. '
             '(agency, network, station, location) for stations nuts.')
    kinds = List.T(
        String.T(),
        help='Available content types in selection.')
    total_size = Int.T(
        help='Aggregated file size of files is selection.')
    counts = Dict.T(
        String.T(), Dict.T(Tuple.T(content_t=String.T()), Int.T()),
        help='Breakdown of how many nuts of any content type and code '
             'sequence are available in selection, ``counts[kind][codes]``.')
    time_spans = Dict.T(
        String.T(), Tuple.T(content_t=Timestamp.T()),
        help='Time spans by content type.')
    sources = List.T(
        String.T(),
        help='Descriptions of attached sources.')
    operators = List.T(
        String.T(),
        help='Descriptions of attached operators.')

    def __str__(self):
        kind_counts = dict(
            (kind, sum(self.counts[kind].values())) for kind in self.kinds)

        scodes = model.codes_to_str_abbreviated(self.codes)

        ssources = '<none>' if not self.sources else '\n' + '\n'.join(
            '  ' + s for s in self.sources)

        soperators = '<none>' if not self.operators else '\n' + '\n'.join(
            '  ' + s for s in self.operators)

        def stime(t):
            return util.tts(t) if t is not None and t not in (
                model.g_tmin, model.g_tmax) else '<none>'

        def stable(rows):
            ns = [max(len(w) for w in col) for col in zip(*rows)]
            return '\n'.join(
                ' '.join(w.ljust(n) for n, w in zip(ns, row))
                for row in rows)

        def indent(s):
            return '\n'.join('  '+line for line in s.splitlines())

        stspans = '<none>' if not self.kinds else '\n' + indent(stable([(
            kind + ':',
            str(kind_counts[kind]),
            stime(self.time_spans[kind][0]),
            '-',
            stime(self.time_spans[kind][1])) for kind in sorted(self.kinds)]))

        s = '''
Number of files:               %i
Total size of known files:     %s
Number of index nuts:          %i
Available content kinds:       %s
Available codes:               %s
Sources:                       %s
Operators:                     %s''' % (
            self.nfiles,
            util.human_bytesize(self.total_size),
            self.nnuts,
            stspans, scodes, ssources, soperators)

        return s.lstrip()


__all__ = [
    'Squirrel',
    'SquirrelStats',
]



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num
import logging

from pyrockoeost import moment_tensor as mt
from pyrockoeost.guts import Float, String, Timestamp, Int
from pyrockoeost.model import Location
from pyrockoeost.modelling import okada_ext
from pyrockoeost.util import get_threadpool_limits

guts_prefix = 'modelling'

logger = logging.getLogger(__name__)

d2r = num.pi/180.
r2d = 180./num.pi
km = 1e3


class AnalyticalSource(Location):
    '''
    Base class for analytical source models.
    '''

    name = String.T(
        optional=True,
        default='')

    time = Timestamp.T(
        default=0.,
        help='Source origin time',
        optional=True)

    vr = Float.T(
        default=0.,
        help='Rupture velocity [m/s]',
        optional=True)

    @property
    def northing(self):
        return self.north_shift

    @property
    def easting(self):
        return self.east_shift


class AnalyticalRectangularSource(AnalyticalSource):
    '''
    Rectangular analytical source model.

    Coordinates on the source plane are with respect to the origin point given
    by `(lat, lon, east_shift, north_shift, depth)`.
    '''

    strike = Float.T(
        default=0.0,
        help='Strike direction in [deg], measured clockwise from north.')

    dip = Float.T(
        default=90.0,
        help='Dip angle in [deg], measured downward from horizontal.')

    rake = Float.T(
        default=0.0,
        help='Rake angle in [deg], measured counter-clockwise from '
             'right-horizontal in on-plane view.')

    al1 = Float.T(
        default=0.,
        help='Left edge source plane coordinate [m].')

    al2 = Float.T(
        default=0.,
        help='Right edge source plane coordinate [m].')

    aw1 = Float.T(
        default=0.,
        help='Lower edge source plane coordinate [m].')

    aw2 = Float.T(
        default=0.,
        help='Upper edge source plane coordinate [m].')

    slip = Float.T(
        default=0.,
        help='Slip on the rectangular source area [m].',
        optional=True)

    @property
    def length(self):
        return abs(-self.al1 + self.al2)

    @property
    def width(self):
        return abs(-self.aw1 + self.aw2)

    @property
    def area(self):
        return self.width * self.length


class OkadaSource(AnalyticalRectangularSource):
    '''
    Rectangular Okada source model.
    '''

    opening = Float.T(
        default=0.,
        help='Opening of the plane in [m].',
        optional=True)

    poisson__ = Float.T(
        default=0.25,
        help='Poisson ratio :math:`\\nu`. '
             'The Poisson ratio :math:`\\nu`. If set to ``None``, calculated '
             "from the Lame' parameters :math:`\\lambda` and :math:`\\mu` "
             'using :math:`\\nu = \\frac{\\lambda}{2(\\lambda + \\mu)}` (e.g. '
             'Mueller 2007).',
        optional=True)

    lamb__ = Float.T(
        help='First Lame parameter :math:`\\lambda` [Pa]. '
             'If set to ``None``, it is computed from Poisson ratio '
             ':math:`\\nu` and shear modulus :math:`\\mu`. **Important:** We '
             'assume a perfect elastic solid with :math:`K=\\frac{5}{3}\\mu`. '
             'Through :math:`\\nu = \\frac{\\lambda}{2(\\lambda + \\mu)}` '
             'this leads to :math:`\\lambda = \\frac{2 \\mu \\nu}{1-2\\nu}`.',
        optional=True)

    shearmod__ = Float.T(
        default=32.0e9,
        help='Shear modulus :math:`\\mu` [Pa]. '
             'If set to ``None``, it is computed from poisson ratio. '
             '**Important:** We assume a perfect elastic solid with '
             ':math:`K=\\frac{5}{3}\\mu`. Through '
             ':math:`\\mu = \\frac{3K(1-2\\nu)}{2(1+\\nu)}` this leads to '
             ':math:`\\mu = \\frac{8(1+\\nu)}{1-2\\nu}`.',
        optional=True)

    @property
    def poisson(self):
        if self.poisson__ is not None:
            return self.poisson__

        if self.shearmod__ is None or self.lamb__ is None:
            raise ValueError('Shearmod and lambda are needed')

        return (self.lamb__) / (2. * (self.lamb__ + self.shearmod__))

    @poisson.setter
    def poisson(self, poisson):
        self.poisson__ = poisson

    @property
    def lamb(self):

        if self.lamb__ is not None:
            return self.lamb__

        if self.shearmod__ is None or self.poisson__ is None:
            raise ValueError('Shearmod and poisson ratio are needed')

        return (
            2. * self.poisson__ * self.shearmod__) / (1. - 2. * self.poisson__)

    @lamb.setter
    def lamb(self, lamb):
        self.lamb__ = lamb

    @property
    def shearmod(self):

        if self.shearmod__ is not None:
            return self.shearmod__

        if self.poisson__ is None:
            raise ValueError('Poisson ratio is needed')

        return (8. * (1. + self.poisson__)) / (1. - 2. * self.poisson__)

    @shearmod.setter
    def shearmod(self, shearmod):
        self.shearmod__ = shearmod

    @property
    def seismic_moment(self):
        '''
        Scalar Seismic moment :math:`M_0`.

        Code copied from Kite. It disregards the opening (as for now).
        We assume :math:`M_0 = mu A D`.

        .. important ::

            We assume a perfect elastic solid with :math:`K=\\frac{5}{3}\\mu`.

            Through :math:`\\mu = \\frac{3K(1-2\\nu)}{2(1+\\nu)}` this leads to
            :math:`\\mu = \\frac{8(1+\\nu)}{1-2\\nu}`.

        :return:
            Seismic moment release.
        :rtype:
            float
        '''

        mu = self.shearmod

        disl = 0.
        if self.slip:
            disl = self.slip
        if self.opening:
            disl = (disl**2 + self.opening**2)**.5

        return mu * self.area * disl

    @property
    def moment_magnitude(self):
        '''
        Moment magnitude :math:`M_\\mathrm{w}` from seismic moment.

        We assume :math:`M_\\mathrm{w} = {\\frac{2}{3}}\\log_{10}(M_0) - 10.7`.

        :returns:
            Moment magnitude.
        :rtype:
            float
        '''
        return mt.moment_to_magnitude(self.seismic_moment)

    def source_patch(self):
        '''
        Get source location and geometry array for okada_ext.okada input.

        The values are defined according to Okada (1992).

        :return:
            Source data as input for okada_ext.okada. The order is
            northing [m], easting [m], depth [m], strike [deg], dip [deg],
            al1 [m], al2 [m], aw1 [m], aw2 [m].
        :rtype:
            :py:class:`~numpy.ndarray`: ``(9, )``
        '''
        return num.array([
            self.northing,
            self.easting,
            self.depth,
            self.strike,
            self.dip,
            self.al1,
            self.al2,
            self.aw1,
            self.aw2])

    def source_disloc(self):
        '''
        Get source dislocation array for okada_ext.okada input.

        The given slip is splitted into a strike and an updip part based on the
        source rake.

        :return:
            Source dislocation data as input for okada_ext.okada. The order is
            dislocation in strike [m], dislocation updip [m], opening [m].
        :rtype:
            :py:class:`~numpy.ndarray`: ``(3, )``
        '''
        return num.array([
            num.cos(self.rake * d2r) * self.slip,
            num.sin(self.rake * d2r) * self.slip,
            self.opening])

    def discretize(self, nlength, nwidth, *args, **kwargs):
        '''
        Discretize fault into rectilinear grid of fault patches.

        Fault orientation, slip and elastic parameters are passed to the
        sub-faults unchanged.

        :param nlength:
            Number of patches in strike direction.
        :type nlength:
            int

        :param nwidth:
            Number of patches in down-dip direction.
        :type nwidth:
            int

        :return:
            Discrete fault patches.
        :rtype:
            list of :py:class:`~pyrockoeost.modelling.okada.OkadaPatch`
        '''
        assert nlength > 0
        assert nwidth > 0

        il = num.repeat(num.arange(nlength), nwidth)
        iw = num.tile(num.arange(nwidth), nlength)

        patch_length = self.length / nlength
        patch_width = self.width / nwidth

        al1 = -patch_length / 2.
        al2 = patch_length / 2.
        aw1 = -patch_width / 2.
        aw2 = patch_width / 2.

        source_points = num.zeros((nlength * nwidth, 3))
        source_points[:, 0] = il * patch_length + patch_length / 2.
        source_points[:, 1] = iw * patch_width + patch_width / 2.

        source_points[:, 0] += self.al1
        source_points[:, 1] -= self.aw2

        rotmat = mt.euler_to_matrix(self.dip*d2r, self.strike*d2r, 0.)

        source_points_rot = num.dot(rotmat.T, source_points.T).T
        source_points_rot[:, 0] += self.northing
        source_points_rot[:, 1] += self.easting
        source_points_rot[:, 2] += self.depth

        kwargs = {
            prop: getattr(self, prop) for prop in self.T.propnames
            if prop not in [
                'north_shift', 'east_shift', 'depth',
                'al1', 'al2', 'aw1', 'aw2']}

        return (
            [OkadaPatch(
                parent=self,
                ix=src_point[0],
                iy=src_point[1],
                north_shift=coord[0],
                east_shift=coord[1],
                depth=coord[2],
                al1=al1, al2=al2, aw1=aw1, aw2=aw2, **kwargs)
             for src_point, coord in zip(source_points, source_points_rot)],
            source_points)


class OkadaPatch(OkadaSource):

    '''
    Okada source with additional 2D indexes for bookkeeping.
    '''

    ix = Int.T(help='Relative index of the patch in x')
    iy = Int.T(help='Relative index of the patch in y')

    def __init__(self, parent=None, *args, **kwargs):
        OkadaSource.__init__(self, *args, **kwargs)
        self.parent = parent


def make_okada_coefficient_matrix(
        source_patches_list,
        pure_shear=False,
        rotate_sdn=True,
        nthreads=1, variant='normal'):

    '''
    Build coefficient matrix for given fault patches.

    The boundary element method (BEM) for a discretized fault and the
    determination of the slip distribution :math:`\\Delta u` from stress drop
    :math:`\\Delta \\sigma` is based on
    :math:`\\Delta \\sigma = \\mathbf{C} \\cdot \\Delta u`. Here the
    coefficient matrix :math:`\\mathbf{C}` is built, based on the displacements
    from Okada's solution (Okada, 1992) and their partial derivatives.

    :param source_patches_list:
        Source patches, to be used in BEM.
    :type source_patches_list:
        list of :py:class:`~pyrockoeost.modelling.okada.OkadaSource`.

    :param pure_shear:
        If ``True``, only shear forces are taken into account.
    :type pure_shear:
        optional, bool

    :param rotate_sdn:
        If ``True``, rotate to strike, dip, normal.
    :type rotate_sdn:
        optional, bool

    :param nthreads:
        Number of threads.
    :type nthreads:
        optional, int

    :return:
        Coefficient matrix for all source combinations.
    :rtype:
        :py:class:`~numpy.ndarray`:
        ``(len(source_patches_list) * 3, len(source_patches_list) * 3)``
    '''

    if variant == 'slow':
        return _make_okada_coefficient_matrix_slow(
            source_patches_list, pure_shear, rotate_sdn, nthreads)

    source_patches = num.array([
        src.source_patch() for src in source_patches_list])
    receiver_coords = source_patches[:, :3].copy()

    npoints = len(source_patches_list)

    if pure_shear:
        n_eq = 2
    else:
        n_eq = 3

    coefmat = num.zeros((npoints * 3, npoints * 3))

    lambda_mean = num.mean([src.lamb for src in source_patches_list])
    mu_mean = num.mean([src.shearmod for src in source_patches_list])

    unit_disl = 1.
    disl_cases = {
        'strikeslip': {
            'slip': unit_disl,
            'opening': 0.,
            'rake': 0.},
        'dipslip': {
            'slip': unit_disl,
            'opening': 0.,
            'rake': 90.},
        'tensileslip': {
            'slip': 0.,
            'opening': unit_disl,
            'rake': 0.}
    }

    diag_ind = [0, 4, 8]
    kron = num.zeros(9)
    kron[diag_ind] = 1.

    if variant == 'normal':
        kron = kron[num.newaxis, num.newaxis, :]
    else:
        kron = kron[num.newaxis, :]

    for idisl, case_type in enumerate([
            'strikeslip', 'dipslip', 'tensileslip'][:n_eq]):
        case = disl_cases[case_type]
        source_disl = num.array([
            case['slip'] * num.cos(case['rake'] * d2r),
            case['slip'] * num.sin(case['rake'] * d2r),
            case['opening']])

        if variant == 'normal':
            results = okada_ext.okada(
                source_patches,
                num.tile(source_disl, npoints).reshape(-1, 3),
                receiver_coords,
                lambda_mean,
                mu_mean,
                nthreads=nthreads,
                rotate_sdn=int(rotate_sdn),
                stack_sources=int(variant != 'normal'))

            eps = 0.5 * (
                results[:, :, 3:] +
                results[:, :, (3, 6, 9, 4, 7, 10, 5, 8, 11)])

            dilatation \
                = eps[:, :, diag_ind].sum(axis=-1)[:, :, num.newaxis]

            stress_sdn = kron*lambda_mean*dilatation + 2.*mu_mean*eps
            coefmat[:, idisl::3] = stress_sdn[:, :, (2, 5, 8)]\
                .reshape(-1, npoints*3).T
        else:
            for isrc, source in enumerate(source_patches):
                results = okada_ext.okada(
                    source.reshape(1, -1),
                    source_disl.reshape(1, -1),
                    receiver_coords,
                    lambda_mean,
                    mu_mean,
                    nthreads=nthreads,
                    rotate_sdn=int(rotate_sdn))

                eps = 0.5 * (
                    results[:, 3:] +
                    results[:, (3, 6, 9, 4, 7, 10, 5, 8, 11)])

                dilatation \
                    = num.sum(eps[:, diag_ind], axis=1)[:, num.newaxis]
                stress_sdn \
                    = kron * lambda_mean * dilatation+2. * mu_mean * eps

                coefmat[:, isrc*3 + idisl] \
                    = stress_sdn[:, (2, 5, 8)].ravel()

    if pure_shear:
        coefmat[2::3, :] = 0.

    return -coefmat / unit_disl


def _make_okada_coefficient_matrix_slow(
        source_patches_list, pure_shear=False, rotate_sdn=True, nthreads=1):

    source_patches = num.array([
        src.source_patch() for src in source_patches_list])
    receiver_coords = source_patches[:, :3].copy()

    npoints = len(source_patches_list)

    if pure_shear:
        n_eq = 2
    else:
        n_eq = 3

    coefmat = num.zeros((npoints * 3, npoints * 3))

    def ned2sdn_rotmat(strike, dip):
        rotmat = mt.euler_to_matrix(
            (dip + 180.) * d2r, strike * d2r, 0.)
        return rotmat

    lambda_mean = num.mean([src.lamb for src in source_patches_list])
    shearmod_mean = num.mean([src.shearmod for src in source_patches_list])

    unit_disl = 1.
    disl_cases = {
        'strikeslip': {
            'slip': unit_disl,
            'opening': 0.,
            'rake': 0.},
        'dipslip': {
            'slip': unit_disl,
            'opening': 0.,
            'rake': 90.},
        'tensileslip': {
            'slip': 0.,
            'opening': unit_disl,
            'rake': 0.}
    }
    for idisl, case_type in enumerate([
            'strikeslip', 'dipslip', 'tensileslip'][:n_eq]):
        case = disl_cases[case_type]
        source_disl = num.array([
            case['slip'] * num.cos(case['rake'] * d2r),
            case['slip'] * num.sin(case['rake'] * d2r),
            case['opening']])

        for isource, source in enumerate(source_patches):
            results = okada_ext.okada(
                source[num.newaxis, :].copy(),
                source_disl[num.newaxis, :].copy(),
                receiver_coords,
                lambda_mean,
                shearmod_mean,
                nthreads=nthreads,
                rotate_sdn=int(rotate_sdn))

            for irec in range(receiver_coords.shape[0]):
                eps = num.zeros((3, 3))
                for m in range(3):
                    for n in range(3):
                        eps[m, n] = 0.5 * (
                            results[irec][m * 3 + n + 3] +
                            results[irec][n * 3 + m + 3])

                stress = num.zeros((3, 3))
                dilatation = num.sum([eps[i, i] for i in range(3)])

                for m, n in zip([0, 0, 0, 1, 1, 2], [0, 1, 2, 1, 2, 2]):
                    if m == n:
                        stress[m, n] = \
                            lambda_mean * \
                            dilatation + \
                            2. * shearmod_mean * \
                            eps[m, n]

                    else:
                        stress[m, n] = \
                            2. * shearmod_mean * \
                            eps[m, n]
                        stress[n, m] = stress[m, n]

                normal = num.array([0., 0., -1.])
                for isig in range(3):
                    tension = num.sum(stress[isig, :] * normal)
                    coefmat[irec * n_eq + isig, isource * n_eq + idisl] = \
                        tension / unit_disl

    return coefmat


def invert_fault_dislocations_bem(
        stress_field,
        coef_mat=None,
        source_list=None,
        pure_shear=False,
        epsilon=None,
        nthreads=1,
        **kwargs):
    '''
    BEM least squares inversion to get fault dislocations given stress field.

    Follows least squares inversion approach by Menke (1989) to calculate
    dislocations on a fault with several segments from a given stress field.
    The coefficient matrix connecting stresses and displacements of the fault
    patches can either be specified by the user (``coef_mat``) or it is
    calculated using the solution of Okada (1992) for a rectangular fault in a
    homogeneous half space (``source_list``).

    :param stress_field:
        Stress change [Pa] for each source patch (as
        ``stress_field[isource, icomponent]`` where isource indexes the source
        patch and ``icomponent`` indexes component, ordered (strike, dip,
        tensile).
    :type stress_field:
        :py:class:`~numpy.ndarray`: ``(nsources, 3)``

    :param coef_mat:
        Coefficient matrix connecting source patch dislocations and the stress
        field.
    :type coef_mat:
        optional, :py:class:`~numpy.ndarray`:
        ``(len(source_list) * 3, len(source_list) * 3)``

    :param source_list:
        Source patches to be used for BEM.
    :type source_list:
        optional, list of
        :py:class:`~pyrockoeost.modelling.okada.OkadaSource`

    :param epsilon:
        If given, values in ``coef_mat`` smaller than ``epsilon`` are set to
        zero.
    :type epsilon:
        optional, float

    :param nthreads:
        Number of threads allowed.
    :type nthreads:
        int

    :return:
        Inverted displacements as ``displacements[isource, icomponent]``
        where isource indexes the source patch and ``icomponent`` indexes
        component, ordered (strike, dip, tensile).
    :rtype:
        :py:class:`~numpy.ndarray`: ``(nsources, 3)``
    '''

    if source_list is not None and coef_mat is None:
        coef_mat = make_okada_coefficient_matrix(
            source_list, pure_shear=pure_shear, nthreads=nthreads,
            **kwargs)

    if epsilon is not None:
        coef_mat[coef_mat < epsilon] = 0.

    idx = num.arange(0, coef_mat.shape[0])
    if pure_shear:
        idx = idx[idx % 3 != 2]

    coef_mat_in = coef_mat[idx, :][:, idx]
    disloc_est = num.zeros(coef_mat.shape[0])

    if stress_field.ndim == 2:
        stress_field = stress_field.ravel()

    threadpool_limits = get_threadpool_limits()

    with threadpool_limits(limits=nthreads, user_api='blas'):
        try:
            disloc_est[idx] = num.linalg.multi_dot([
                num.linalg.inv(num.dot(coef_mat_in.T, coef_mat_in)),
                coef_mat_in.T,
                stress_field[idx]])
        except num.linalg.LinAlgError as e:
            logger.warning('Linear inversion failed!')
            logger.warning(
                'coef_mat: %s\nstress_field: %s',
                coef_mat_in, stress_field[idx])
            raise e
        return disloc_est.reshape(-1, 3)


__all__ = [
    'AnalyticalSource',
    'AnalyticalRectangularSource',
    'OkadaSource',
    'OkadaPatch',
    'make_okada_coefficient_matrix',
    'invert_fault_dislocations_bem']



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from . import base, local, fdsn, catalog

from .base import *  # noqa
from .local import *  # noqa
from .fdsn import *  # noqa
from .catalog import *  # noqa

__all__ = base.__all__ + local.__all__ + fdsn.__all__ + catalog.__all__



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num

from pyrockoeost.guts import \
    Object, Bool, Float, StringChoice, Timestamp, String, List

from pyrockoeost import cake, table, model, geometry
from pyrockoeost.client import fdsn
from pyrockoeost.gui.qt_compat import qw, qc

from pyrockoeost.gui.vtk_util import ScatterPipe
from .. import common

from .base import Element, ElementState

guts_prefix = 'sparrow'


def stations_to_points(stations):
    coords = num.zeros((len(stations), 3))

    for i, s in enumerate(stations):
        coords[i, :] = s.lat, s.lon, -s.elevation

    station_table = table.Table()

    station_table.add_col(('coords', '', ('lat', 'lon', 'depth')), coords)

    return geometry.latlondepth2xyz(
        station_table.get_col('coords'),
        planetradius=cake.earthradius)


class FDSNSiteChoice(StringChoice):
    choices = [key.upper() for key in fdsn.g_site_abbr.keys()]


class StationSelection(Object):
    pass


class FDSNStationSelection(StationSelection):
    site = String.T()
    tmin = Timestamp.T()
    tmax = Timestamp.T()

    def get_stations(self):
        return fdsn.station(
            site=self.site,
            format='text',
            level='channel',
            startbefore=self.tmin,
            endafter=self.tmax
        ).get_pyrockoeost_stations()


class FileStationSelection(StationSelection):
    paths = List.T(String.T())

    def get_stations(self):
        from pyrockoeost.io import stationxml

        stations = []
        for path in self.paths:
            if path.split('.')[-1].lower() in ['xml']:
                stxml = stationxml.load_xml(filename=path)
                stations.extend(stxml.get_pyrockoeost_stations())

            else:
                stations.extend(model.load_stations(path))

        return stations


class StationsState(ElementState):
    visible = Bool.T(default=True)
    size = Float.T(default=5.0)
    station_selection = StationSelection.T(optional=True)

    @classmethod
    def get_name(cls):
        return 'Stations'

    def create(self):
        element = StationsElement()
        return element


class StationsElement(Element):

    def __init__(self):
        Element.__init__(self)
        self._parent = None
        self._pipe = None
        self._state = None
        self._controls = None
        self._points = num.array([])

    def bind_state(self, state):
        Element.bind_state(self, state)
        self.talkie_connect(
            state,
            ['visible', 'size', 'station_selection'],
            self.update)

        self._current_selection = None

    def get_name(self):
        return 'Stations'

    def set_parent(self, parent):
        self._parent = parent
        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])
        self.update()

    def unset_parent(self):
        self.unbind_state()
        if self._parent:
            if self._pipe:
                self._parent.remove_actor(self._pipe.actor)
                self._pipe = None

            if self._controls:
                self._parent.remove_panel(self._controls)
                self._controls = None

            self._parent.update_view()
            self._parent = None

    def update(self, *args):
        state = self._state
        if self._pipe and \
                self._current_selection is not state.station_selection:

            self._parent.remove_actor(self._pipe.actor)
            self._pipe = None

        if self._pipe and not state.visible:
            self._parent.remove_actor(self._pipe.actor)
            self._pipe.set_size(state.size)

        if state.visible:
            if self._current_selection is not state.station_selection:
                stations = state.station_selection.get_stations()
                points = stations_to_points(stations)
                self._pipe = ScatterPipe(points)
                self._parent.add_actor(self._pipe.actor)
            elif self._pipe:
                self._parent.add_actor(self._pipe.actor)

            if self._pipe:
                self._pipe.set_size(state.size)

        self._parent.update_view()
        self._current_selection = state.station_selection

    def open_file_load_dialog(self):
        caption = 'Select one or more files to open'

        fns, _ = qw.QFileDialog.getOpenFileNames(
            self._parent, caption, options=common.qfiledialog_options)

        self._state.station_selection = FileStationSelection(
            paths=[str(fn) for fn in fns])

    def open_fdsn_load_dialog(self):
        dialog = qw.QDialog(self._parent)
        dialog.setWindowTitle('Get stations from FDSN web service')

        layout = qw.QHBoxLayout(dialog)

        layout.addWidget(qw.QLabel('Site'))

        sites = [key.upper() for key in fdsn.g_site_abbr.keys()]

        cb = qw.QComboBox()
        for i, s in enumerate(sites):
            cb.insertItem(i, s)

        layout.addWidget(cb)

        pb = qw.QPushButton('Cancel')
        pb.clicked.connect(dialog.reject)
        layout.addWidget(pb)

        pb = qw.QPushButton('Ok')
        pb.clicked.connect(dialog.accept)
        layout.addWidget(pb)

        dialog.exec_()

        site = str(cb.currentText()).lower()

        vstate = self._parent.state

        if dialog.result() == qw.QDialog.Accepted:
            self._state.station_selection = FDSNStationSelection(
                site=site,
                tmin=vstate.tmin,
                tmax=vstate.tmax)

    def _get_controls(self):
        if not self._controls:
            from ..state import state_bind_slider

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            layout.addWidget(qw.QLabel('Size'), 0, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(10)
            slider.setSingleStep(1)
            slider.setPageStep(1)
            layout.addWidget(slider, 0, 1)
            state_bind_slider(self, self._state, 'size', slider)

            lab = qw.QLabel('Load from:')
            pb_file = qw.QPushButton('File')
            pb_fdsn = qw.QPushButton('FDSN')

            layout.addWidget(lab, 1, 0)
            layout.addWidget(pb_file, 1, 1)
            layout.addWidget(pb_fdsn, 1, 2)

            pb_file.clicked.connect(self.open_file_load_dialog)
            pb_fdsn.clicked.connect(self.open_fdsn_load_dialog)

            layout.addWidget(qw.QFrame(), 2, 0, 1, 3)

            self._controls = frame

        return self._controls


__all__ = [
    'StationsElement',
    'StationsState',
]



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num

from pyrockoeost.guts import Object, Int

from pyrockoeost import icosphere, geometry as g

guts_prefix = 'sparrow'

and_ = num.logical_and


class Level(object):
    def __init__(self, vertices, faces):
        self.vertices = vertices
        self.faces = faces
        self.next = None
        self.nvertices = vertices.shape[0]
        self.nfaces = faces.shape[0]
        self.prepare_projections()

    def points_in_face(self, iface, points):
        uv = g.triangle_barycentric_coordinates(
            self.bary_trans[iface, :, :],
            self.vertices[self.faces[iface, 0]],
            points)

        return and_(
            and_(uv[:, 0] >= 0., uv[:, 1] >= 0.), uv[:, 0] + uv[:, 1] <= 1.)

    def points_to_face(self, points, icandidates=None):

        if icandidates is None:
            icandidates = num.arange(self.nfaces)

        npoints = points.shape[0]
        ifaces = num.zeros(npoints, dtype=int)
        points_flat_all = num.zeros((npoints, 3))
        for icandidate in icandidates:
            points_flat, mask = self.project(icandidate, points)
            ipoints = num.where(num.logical_and(
                self.points_in_face(icandidate, points_flat),
                mask))[0]

            if self.next:
                icandidates_child = num.where(
                    self.next.parents == icandidate)[0]
                ifaces[ipoints], points_flat_all[ipoints, :] \
                    = self.next.points_to_face(
                        points[ipoints], icandidates_child)

            else:
                ifaces[ipoints] = icandidate
                points_flat_all[ipoints, :] = points_flat[ipoints, :]

        return ifaces, points_flat_all

    def project(self, iface, points):
        denom = g.vdot(points, self.aplanes[iface, :])
        mask = denom > 0.0
        points = points.copy()
        points[mask, :] /= denom[mask, num.newaxis]
        points[num.logical_not(mask), :] = 0.0
        return points, mask

    def prepare_projections(self):
        planes = g.triangle_planes(
            self.vertices, self.faces)

        planes /= g.vdot(
            planes,
            self.vertices[self.faces[:, 0]])[:, num.newaxis]

        self.aplanes = planes

        self.bary_trans = g.triangle_barycentric_transforms(
            self.vertices, self.faces)


class HiMesh(Object):

    order = Int.T(default=0)

    def __init__(self, **kwargs):
        Object.__init__(self, **kwargs)
        self.levels = self.make_levels()

    def make_levels(self):
        # root = None
        level = None
        levels = []
        for vertices, faces in icosphere.iter_icospheres(
                self.order, inflate=False):

            new = Level(vertices, faces)
            if not levels:
                new.parents = None
            else:
                new.parents = g.refine_triangle_parents(faces.shape[0])
                level.next = new

            level = new
            levels.append(new)

        return levels

    def get_vertices(self):
        return g.normalize(self.levels[-1].vertices)

    def get_faces(self):
        return self.levels[-1].faces

    def points_to_faces(self, points):
        return self.levels[0].points_to_face(points)[0]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging

from pyrockoeost.guts import Object, Int

logger = logging.getLogger('psq.cache')


class ContentCacheStats(Object):
    '''
    Information about cache state.
    '''
    nentries = Int.T(
        help='Number of items in the cache.')
    naccessors = Int.T(
        help='Number of accessors currently holding references to cache '
             'items.')


class ContentCache(object):

    '''
    Simple memory cache for file contents.

    Squirrel manages data in small entities: nuts. Only the meta-data for each
    nut is stored in the database, content data has to be read from file. This
    cache helps to speed up data access for typical seismological access
    patterns.

    Content data for stations, channels and instrument responses is small in
    size but slow to parse so it makes sense to cache these indefinitely once
    read. Also, it is usually inefficient to read a single station from a
    station file, so it is better to cache the contents of the complete file
    even if only one station is requested (it is likely that other stations
    from that file will be used anyway).

    Content data for waveforms is large in size and we usually want to free the
    memory allocated for them after processing. Typical processing schemes
    require batches of waveforms to be available together (e.g.
    cross-correlations between pairs of stations) and there may be overlap
    between successive batches (e.g. sliding window processing schemes).

    This cache implementation uses named accessors and batch window counting
    for flexible content caching. Loaded contents are held in memory as long as
    an accessor is holding a reference to it. For each accessor a batch counter
    is maintained, which starts at 0 and is incremented using calls to
    :py:meth:`advance_accessor`. Content accesses are tracked with calls to
    :py:meth:`get`, which sets a "last access" attribute on the cached item to
    the current value of the batch counter (each accessor has its own last
    access attribute on the items it uses). References to items which have
    not been accessed during the latest batch by the accessor in question are
    released during :py:meth:`advance_accessor`. :py:meth:`put` inserts new
    items into the cache. :py:meth:`has` checks if there already is content
    cached for a given item. To remove all references held by a given accessor,
    :py:meth:`clear_accessor` can be called.

    **Example usage**

    For meta-data content to be cached indefinitely, no calls to
    :py:meth:`advance_accessor` or :py:meth:`clear_accessor` should be made.
    For waveform content one would call :py:meth:`advance_accessor` after each
    move of a sliding window or :py:meth:`clear_accessor` after each processed
    event. For a process requiring data from two independent positions of
    extraction, e.g. for cross-correlations between all possible pairs of a set
    of events, two separate accessor names could be used.
    '''

    def __init__(self):
        self._entries = {}
        self._accessor_ticks = {}

    def _prune_outdated(self, path, segment, nut_mtime):
        try:
            cache_mtime = self._entries[path, segment][0]
        except KeyError:
            return

        if cache_mtime != nut_mtime:
            logger.debug('Forgetting (outdated): %s %s' % (path, segment))
            del self._entries[path, segment]

    def put(self, nut):
        '''
        Insert a new/updated item into cache.

        :param nut:
            Content item with attached data object.
        :type nut:
            :py:class:`~pyrockoeost.squirrel.model.Nut`
        '''
        path, segment, element, mtime = nut.key
        self._prune_outdated(path, segment, nut.file_mtime)

        if (path, segment) not in self._entries:
            self._entries[path, segment] = nut.file_mtime, {}, {}

        self._entries[path, segment][1][element] = nut

    def get(self, nut, accessor='default', model='squirrel'):
        '''
        Get a content item and track its access.

        :param nut:
            Content item.
        :type nut:
            :py:class:`~pyrockoeost.squirrel.model.Nut`

        :param accessor:
            Name of accessing consumer. Giving a new name initializes a new
            accessor.
        :type accessor:
            str

        :returns:
            Content data object
        '''
        path, segment, element, mtime = nut.key
        entry = self._entries[path, segment]

        if accessor not in self._accessor_ticks:
            self._accessor_ticks[accessor] = 0

        entry[2][accessor] = self._accessor_ticks[accessor]
        el = entry[1][element]

        if model == 'squirrel':
            return el.content
        elif model.endswith('+'):
            return el.content, el.raw_content[model[:-1]]
        else:
            return el.raw_content[model]

    def has(self, nut):
        '''
        Check if item's content is currently in cache.

        :param nut:
            Content item.
        :type nut:
            :py:class:`~pyrockoeost.squirrel.model.Nut`

        :returns:
            :py:class:`bool`

        '''
        path, segment, element, nut_mtime = nut.key

        try:
            entry = self._entries[path, segment]
            cache_mtime = entry[0]
            entry[1][element]
        except KeyError:
            return False

        return cache_mtime == nut_mtime

    def advance_accessor(self, accessor='default'):
        '''
        Increment batch counter of an accessor.

        :param accessor:
            Name of accessing consumer. Giving a new name initializes a new
            accessor.
        :type accessor:
            str
        '''
        if accessor not in self._accessor_ticks:
            self._accessor_ticks[accessor] = 0

        ta = self._accessor_ticks[accessor]

        delete = []
        for path_segment, entry in self._entries.items():
            t = entry[2].get(accessor, ta)
            if t < ta:
                del entry[2][accessor]
                if not entry[2]:
                    delete.append(path_segment)

        for path_segment in delete:
            logger.debug('Forgetting (advance): %s %s' % path_segment)
            del self._entries[path_segment]

        self._accessor_ticks[accessor] += 1

    def clear_accessor(self, accessor='default'):
        '''
        Clear all references held by an accessor.

        :param accessor:
            Name of accessing consumer.
        :type accessor:
            str
        '''
        delete = []
        for path_segment, entry in self._entries.items():
            entry[2].pop(accessor, None)
            if not entry[2]:
                delete.append(path_segment)

        for path_segment in delete:
            logger.debug('Forgetting (clear): %s %s' % path_segment)
            del self._entries[path_segment]

        del self._accessor_ticks[accessor]

    def clear(self):
        '''
        Empty the cache.
        '''
        for accessor in list(self._accessor_ticks.keys()):
            self.clear_accessor(accessor)

        self._entries = {}
        self._accessor_ticks = {}

    def get_stats(self):
        '''
        Get information about cache state.

        :returns: :py:class:`ContentCacheStats` object.
        '''
        return ContentCacheStats(
            nentries=len(self._entries),
            naccessors=len(self._accessor_ticks))



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from pyrockoeost.guts import Object, Timestamp, List

from ..model import CodesNSLCE


guts_prefix = 'squirrel'


class Constraint(Object):

    '''
    Used by some data-sources to grow or join locally mirrored data selections.

    Squirrel data-sources typically try to mirror only a subset of the remotely
    available data. This subset may need to be grown or updated when data from
    other time intervals or from locations outside the initially requested
    region is requested. This class helps in the involved bookeeping.

    The current implementation only supports a time interval selection with a
    single time span but more sophisticated constraints, including e.g.
    location boxes could be thought of.
    '''

    tmin = Timestamp.T(optional=True)
    tmax = Timestamp.T(optional=True)
    codes = List.T(CodesNSLCE.T(), optional=True)

    def __init__(self, **kwargs):
        codes = kwargs.pop('codes', None)

        if codes is None:
            pass
        elif not isinstance(codes, list):
            codes = [CodesNSLCE(codes)]
        else:
            codes = [CodesNSLCE(sc) for sc in codes]

        Object.__init__(self, codes=codes, **kwargs)

    def contains(self, constraint):
        '''
        Check if the constraint completely includes a more restrictive one.

        :param constraint:
            Other constraint.
        :type constraint:
            :py:class:`Constraint`
        '''

        if self.tmin is not None and constraint.tmin is not None:
            b1 = self.tmin <= constraint.tmin
        elif self.tmin is None:
            b1 = True
        else:
            b1 = False

        if self.tmax is not None and constraint.tmax is not None:
            b2 = constraint.tmax <= self.tmax
        elif self.tmax is None:
            b2 = True
        else:
            b2 = False

        return b1 and b2

    def expand(self, constraint):
        '''
        Widen constraint to include another given constraint.

        :param constraint:
            Other constraint.
        :type constraint:
            :py:class:`Constraint`

        Update is done in-place.
        '''

        if constraint.tmin is None or self.tmin is None:
            self.tmin = None
        else:
            self.tmin = min(constraint.tmin, self.tmin)

        if constraint.tmax is None or self.tmax is None:
            self.tmax = None
        else:
            self.tmax = max(constraint.tmax, self.tmax)


class Source(Object):

    '''
    Base class for Squirrel data-sources.

    Data-sources can be attached to a Squirrel instance to allow transparent
    access to remote (or otherwise generated) resources, e.g. through FDSN web
    services (:py:class:`~pyrockoeost.squirrel.client.fdsn.FDSNSource`) or online
    event catalogs (:py:class:`~pyrockoeost.client.catalog.CatalogSource`).

    Derived classes implement the details of querying, caching, updating and
    bookkeeping of the accessed data.
    '''

    def update_channel_inventory(self, squirrel, constraint):
        '''
        Let local inventory be up-to-date with remote for a given constraint.
        '''

        pass

    def update_event_inventory(self, squirrel, constraint):
        '''
        Let local inventory be up-to-date with remote for a given constraint.
        '''

        pass

    def update_waveform_promises(self, squirrel, constraint):
        '''
        Let local inventory be up-to-date with remote for a given constraint.
        '''

        pass

    def remove_waveform_promises(self, squirrel, from_database='selection'):
        '''
        Remove waveform promises from live selection or global database.

        :param from_database:
            Remove from live selection ``'selection'`` or global database
            ``'global'``.
        '''

        pass

    def update_response_inventory(self, squirrel, constraint):
        '''
        Let local inventory be up-to-date with remote for a given constraint.
        '''

        pass


__all__ = [
    'Source',
    'Constraint',
]
